from __future__ import annotations

import abc
import functools as ft
import re
from collections.abc import Iterable, Mapping, Sequence
from typing import TYPE_CHECKING, Any, Final, TypeAlias, cast

import attrs
import more_itertools as itx
from bids.layout import BIDSLayout, Query
from bids.layout.models import BIDSFile
from typing_extensions import Self, override

from snakebids.exceptions import ConfigError, PybidsError
from snakebids.types import FilterMap, FilterValue, InputConfig

CompiledFilter: TypeAlias = "Mapping[str, Sequence[str | Query]]"


class PostFilter:
    """Filters to apply after indexing, typically derived from the CLI.

    Currently used for supporting ``--[exclude-]participant-label``
    """

    def __init__(self):
        self.inclusions: dict[str, Sequence[str] | str] = {}
        self.exclusions: dict[str, Sequence[str] | str] = {}

    def add_filter(
        self,
        key: str,
        inclusions: Iterable[str] | str | None,
        exclusions: Iterable[str] | str | None,
    ):
        """Add entity filter based on inclusion or exclusion criteria.

        Converts a list of values to include or exclude into Pybids compatible filters.
        Exclusion filters are appropriately formatted as regex. Raises an exception if
        both include and exclude are stipulated.

        PostFilter is modified in-place.

        Parameters
        ----------
        key
            Name of entity to be filtered
        inclusions
            Values to include, values not found in this list will be excluded, by
            default ``None``
        exclusions
            Values to exclude, only values not found in this list will be included, by
            default ``None``

        Raises
        ------
        ValueError
            Raised if both include and exclude values are stipulated.
        """
        if inclusions is not None:
            self.inclusions[key] = list(itx.always_iterable(inclusions))
        if exclusions is not None:
            self.exclusions[key] = self._format_exclusions(exclusions)

    def _format_exclusions(self, exclusions: Iterable[str] | str):
        # if multiple items to exclude, combine with with item1|item2|...
        exclude_string = "|".join(
            re.escape(label) for label in itx.always_iterable(exclusions)
        )
        # regex to exclude subjects
        return [f"^((?!({exclude_string})$).*)$"]


@attrs.define(slots=False)
class UnifiedFilter:
    """Manages component level and post filters."""

    component: InputConfig
    """The Component configuration defining the filters"""

    postfilters: PostFilter
    """Filters to be applied after collecting and parsing the data

    Currently only used to implement --[exclude-]participant-label, but in the future,
    may implement other such CLI args. Unlike configuration-defined filters, these
    filters apply after the dataset is indexed and queried. Thus, if a filter is set
    to an empty list, a complete, albeit empty, component may be found. This is akin to
    calling ``BidsComponent.filter`` after running ``generate_inputs``.

    For performance purposes, non-empty post-filters are applied via ``pybids.get()``
    """

    @classmethod
    def from_filter_dict(
        cls,
        filters: Mapping[str, str | bool | Sequence[str | bool]],
        postfilter: PostFilter | None = None,
    ) -> Self:
        """Patch together a UnifiedFilter based on a basic filter dict.

        Intended primarily for use in testing
        """
        wildcards: list[str] = []
        if postfilter is not None:
            wildcards.extend(postfilter.inclusions)
            wildcards.extend(postfilter.exclusions)
        return cls(
            {"filters": filters, "wildcards": wildcards}, postfilter or PostFilter()
        )

    def _has_empty_list(self, items: Iterable[Any]):
        """Check if any of the lists within iterable are empty."""
        return any(
            itx.ilen(itx.always_iterable(item, base_type=(str, dict))) == 0
            for item in items
        )

    def _has_overlap(self, key: str):
        """Check if filter key is a wildcard and not already a prefilter."""
        return key not in self.prefilters and key in self.component.get("wildcards", [])

    @ft.cached_property
    def prefilters(self) -> FilterMap:
        """Filters defined in the component configuration and applied via pybids.

        Unlike postfilters, a prefilter set to an empty list will result in no valid
        paths found, resulting in a blank (missing) component.
        """
        filters = dict(self.component.get("filters", {}))
        # Silently remove "regex_search". This value has been blocked by a bug for the
        # since version 0.6, and even before, never fully worked properly (e.g. would
        # break if combined with --exclude-participant-label)
        filters.pop("regex_search", None)
        return filters

    @ft.cached_property
    def get(self) -> CompiledFilter:
        """The combination of pre- and post- filters for indexing pybids via ``.get()``.

        Includes pre-filters not annotated for regex querying and all inclusion
        post-filters. Empty post-filters are replaced with Query.ANY. This allows valid
        paths to be found and processed later. Post-filters are not applied when an
        equivalent prefilter is present

        Raises
        ------
        FilterSpecError
            When filter configuration is invalidly specified.
        """
        result = dict(_compile_filters(self.prefilters, with_regex=False))
        postfilters = self.postfilters.inclusions
        for key in self.postfilters.inclusions:
            if self._has_overlap(key):
                # if empty list filter, ensure the entity filtered is present
                result[key] = (
                    postfilters[key]
                    if itx.ilen(itx.always_iterable(postfilters[key]))
                    else [Query.ANY]
                )
        return result

    @ft.cached_property
    def search(self) -> CompiledFilter:
        """Pre-filters for indexing pybids via ``.get(regex_search=True)``.

        As with :prop:`UnifiedFilter.get`, but only prefilters labelled for regex
        matching using ``search:`` or ``match:``.

        Raises
        ------
        FilterSpecError
            When filter configuration is invalidly specified.
        """
        return dict(_compile_filters(self.prefilters, with_regex=True))

    @property
    def post_exclusions(self) -> dict[str, Sequence[str] | str]:
        """Dictionary of all post-exclusion filters."""
        return {
            key: val
            for key, val in self.postfilters.exclusions.items()
            if self._has_overlap(key)
        }

    @property
    def without_bools(self) -> Mapping[str, str | Sequence[str]]:
        """Check and typeguard to ensure filters do not contain booleans."""
        for key, val in self.get.items():
            if any(isinstance(v, Query) for v in itx.always_iterable(val)):
                msg = (
                    "Boolean filters in items with custom paths are not supported; in "
                    f"component='{key}'"
                )
                raise ValueError(msg)
        return cast("Mapping[str, str | Sequence[str]]", self.get)

    @property
    def has_empty_prefilter(self) -> bool:
        """Returns True if even one prefilter is empty."""
        return self._has_empty_list(self.prefilters.values())

    @property
    def has_empty_postfilter(self) -> bool:
        """Returns True if even one postfilter is empty."""
        return self._has_empty_list(
            filt
            for name, filt in self.postfilters.inclusions.items()
            if self._has_overlap(name)
        )


def get_matching_files(
    bids_layout: BIDSLayout,
    filters: UnifiedFilter,
) -> Iterable[BIDSFile]:
    """Query pybids layout based on provided filters.

    Supports a combination of regular and regex querying.

    Raises
    ------
    FilterSpecError
        When filter configuration is invalidly specified.
    PybidsError
        When pybids raises an error within BIDSLayout.get()
    """
    if filters.has_empty_prefilter:
        return []
    try:
        get = bids_layout.get(
            regex_search=False,
            **filters.get,
        )
        search = (
            set(bids_layout.get(regex_search=True, **filters.search))
            if filters.search
            else None
        )
    except AttributeError as err:
        msg = (
            "Pybids has encountered a problem that Snakebids cannot handle. This "
            "may indicate a missing or invalid dataset_description.json for this "
            "dataset."
        )
        raise PybidsError(msg) from err

    if search is not None:
        return [p for p in get if p in search]
    return get


@attrs.define
class FilterSpecError(Exception, abc.ABC):
    entity: str
    value: FilterValue

    requirement: Final[str] = attrs.field(
        default="Must have exactly one of {'get', 'match', search'}.", init=False
    )

    @abc.abstractmethod
    def get_config_error(self, component_name: str) -> ConfigError:
        """Return ConfigError with class-specific message."""
        ...


@attrs.define
class _TooFewKeysError(FilterSpecError):
    """Exception raised when filter specified as dictionary without keys."""

    @override
    def get_config_error(self, component_name: str) -> ConfigError:
        msg = (
            f"Filter '{self.entity}' for component '{component_name}' was specified as "
            f"a dict but was not given any keys. {self.requirement} Got: {self.value}"
        )
        return ConfigError(msg)


@attrs.define
class _TooManyKeysError(FilterSpecError):
    """Exception raised when filter specified as dictionary with more than one key."""

    @override
    def get_config_error(self, component_name: str) -> ConfigError:
        msg = (
            f"Filter '{self.entity}' for component '{component_name}' may not have "
            f"more than one key. {self.requirement} Got: {self.value}"
        )
        return ConfigError(msg)


@attrs.define
class _InvalidKeyError(FilterSpecError):
    """Exception raised when filter specified as dictionary with invalid key."""

    @override
    def get_config_error(self, component_name: str) -> ConfigError:
        msg = (
            f"Invalid query method specified for filter '{self.entity}' in component "
            f"'{component_name}'. {self.requirement} Got {self.value}"
        )
        return ConfigError(msg)


def _compile_filters(filters: FilterMap, *, with_regex: bool) -> CompiledFilter:
    """Convert filter configuration into dict consumable by pybids get().

    Raises
    ------
    FilterSpecError
        When filter configuration is invalidly specified.
    """

    def wrap(filt: str, method: str):
        if method == "get":
            return filt
        # use flag group to remove the default case-insensitivity enforced by pybids
        filt = f"(?-i:{filt})"
        if method == "match":
            # pybids only does search, so surround string filters with position anchors
            # to simulate match
            filt = f"^(?:{filt})$"
        return filt

    result: CompiledFilter = {}
    for key, raw_filter in filters.items():
        try:
            filt_type = itx.one(raw_filter.keys(), too_short=TypeError)  # type: ignore
        except ValueError as err:
            raise _TooManyKeysError(key, raw_filter) from err
        except TypeError as err:
            raise _TooFewKeysError(key, raw_filter) from err
        except AttributeError:
            if TYPE_CHECKING:
                assert not isinstance(raw_filter, dict)
            filt = raw_filter
            filt_type = "get"
        else:
            if filt_type not in {"match", "search", "get"}:
                raise _InvalidKeyError(key, raw_filter)
            if TYPE_CHECKING:
                assert isinstance(raw_filter, dict)
            filt = cast("str | bool | Sequence[str | bool]", raw_filter[filt_type])

        # these two must not be simultaneously true or false
        if with_regex == (filt_type == "get"):
            continue

        result[key] = [
            Query.ANY if f is True else Query.NONE if f is False else wrap(f, filt_type)
            for f in itx.always_iterable(filt)
        ]

    return result



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.compat import MutableSliceableSequence, ordereddict
from ruamel.yaml.tag import Tag
from collections.abc import Mapping, MutableSet, Set, Sized

if False: ...
__all__ = [
    "CommentedSeq",
    "CommentedKeySeq",
    "CommentedMap",
    "CommentedOrderedMap",
    "CommentedSet",
    "comment_attrib",
    "merge_attrib",
    "TaggedScalar",
    "C_POST",
    "C_PRE",
    "C_SPLIT_ON_FIRST_BLANK",
    "C_BLANK_LINE_PRESERVE_SPACE",
]
C_POST = ...
C_PRE = ...
C_SPLIT_ON_FIRST_BLANK = ...
C_BLANK_LINE_PRESERVE_SPACE = ...

class IDX:
    def __init__(self) -> None: ...
    def __call__(self) -> Any: ...
    def __str__(self) -> Any: ...

cidx = ...
C_ELEM_EOL = ...
C_KEY_EOL = ...
C_ELEM_PRE = ...
C_ELEM_POST = ...
C_VALUE_PRE = ...
C_KEY_POST = ...
C_TAG_EOL = ...
C_TAG_POST = ...
C_TAG_PRE = ...
C_ANCHOR_EOL = ...
C_ANCHOR_POST = ...
C_ANCHOR_PRE = ...
comment_attrib = ...
format_attrib = ...
line_col_attrib = ...
merge_attrib = ...

class Comment:
    __slots__ = ...
    attrib = ...
    def __init__(self, old: bool = ...) -> None: ...
    def __str__(self) -> str: ...
    def __repr__(self) -> str: ...
    @property
    def items(self) -> Any: ...
    @property
    def end(self) -> Any: ...
    @end.setter
    def end(self, value: Any) -> None: ...
    @property
    def pre(self) -> Any: ...
    @pre.setter
    def pre(self, value: Any) -> None: ...
    def get(self, item: Any, pos: Any) -> Any: ...
    def set(self, item: Any, pos: Any, value: Any) -> Any: ...
    def __contains__(self, x: Any) -> Any: ...

class NotNone: ...

class Format:
    __slots__ = ...
    attrib = ...
    def __init__(self) -> None: ...
    def set_flow_style(self) -> None: ...
    def set_block_style(self) -> None: ...
    def flow_style(self, default: Optional[Any] = ...) -> Any:
        """if default (the flow_style) is None, the flow style tacked on to
        the object explicitly will be taken. If that is None as well the
        default flow style rules the format down the line, or the type
        of the constituent values (simple -> flow, map/list -> block)"""
        ...

    def __repr__(self) -> str: ...

class LineCol:
    """
    line and column information wrt document, values start at zero (0)
    """

    attrib = ...
    def __init__(self) -> None: ...
    def add_kv_line_col(self, key: Any, data: Any) -> None: ...
    def key(self, k: Any) -> Any: ...
    def value(self, k: Any) -> Any: ...
    def item(self, idx: Any) -> Any: ...
    def add_idx_line_col(self, key: Any, data: Any) -> None: ...
    def __repr__(self) -> str: ...

class CommentedBase:
    @property
    def ca(self) -> Any: ...
    def yaml_end_comment_extend(self, comment: Any, clear: bool = ...) -> None: ...
    def yaml_key_comment_extend(
        self, key: Any, comment: Any, clear: bool = ...
    ) -> None: ...
    def yaml_value_comment_extend(
        self, key: Any, comment: Any, clear: bool = ...
    ) -> None: ...
    def yaml_set_start_comment(self, comment: Any, indent: Any = ...) -> None:
        """overwrites any preceding comment lines on an object
        expects comment to be without `#` and possible have multiple lines
        """
        ...

    def yaml_set_comment_before_after_key(
        self,
        key: Any,
        before: Any = ...,
        indent: Any = ...,
        after: Any = ...,
        after_indent: Any = ...,
    ) -> None:
        """
        expects comment (before/after) to be without `#` and possible have multiple lines
        """
        ...

    @property
    def fa(self) -> Any:
        """format attribute

        set_flow_style()/set_block_style()"""
        ...

    def yaml_add_eol_comment(
        self, comment: Any, key: Optional[Any] = ..., column: Optional[Any] = ...
    ) -> None:
        """
        there is a problem as eol comments should start with ' #'
        (but at the beginning of the line the space doesn't have to be before
        the #. The column index is for the # mark
        """
        ...

    @property
    def lc(self) -> Any: ...
    @property
    def anchor(self) -> Any: ...
    def yaml_anchor(self) -> Any: ...
    def yaml_set_anchor(self, value: Any, always_dump: bool = ...) -> None: ...
    @property
    def tag(self) -> Any: ...
    def yaml_set_ctag(self, value: Tag) -> None: ...
    def copy_attributes(self, t: Any, memo: Any = ...) -> Any:
        """
        copies the YAML related attributes, not e.g. .values
        returns target
        """
        ...

class CommentedSeq(MutableSliceableSequence, list, CommentedBase):
    __slots__ = ...
    def __init__(self, *args: Any, **kw: Any) -> None: ...
    def __getsingleitem__(self, idx: Any) -> Any: ...
    def __setsingleitem__(self, idx: Any, value: Any) -> None: ...
    def __delsingleitem__(self, idx: Any = ...) -> Any: ...
    def __len__(self) -> int: ...
    def insert(self, idx: Any, val: Any) -> None:
        """the comments after the insertion have to move forward"""
        ...

    def extend(self, val: Any) -> None: ...
    def __eq__(self, other: Any) -> bool: ...
    def __deepcopy__(self, memo: Any) -> Any: ...
    def __add__(self, other: Any) -> Any: ...
    def sort(self, key: Any = ..., reverse: bool = ...) -> None: ...
    def __repr__(self) -> Any: ...

class CommentedKeySeq(tuple, CommentedBase):
    """This primarily exists to be able to roundtrip keys that are sequences"""

    ...

class CommentedMapView(Sized):
    __slots__ = ...
    def __init__(self, mapping: Any) -> None: ...
    def __len__(self) -> int: ...

class CommentedMapKeysView(CommentedMapView, Set):
    __slots__ = ...
    def __contains__(self, key: Any) -> Any: ...
    def __iter__(self) -> Any: ...

class CommentedMapItemsView(CommentedMapView, Set):
    __slots__ = ...
    def __contains__(self, item: Any) -> Any: ...
    def __iter__(self) -> Any: ...

class CommentedMapValuesView(CommentedMapView):
    __slots__ = ...
    def __contains__(self, value: Any) -> Any: ...
    def __iter__(self) -> Any: ...

class CommentedMap(ordereddict, CommentedBase):
    __slots__ = ...
    def __init__(self, *args: Any, **kw: Any) -> None: ...
    def update(self, *vals: Any, **kw: Any) -> None: ...
    def insert(
        self, pos: Any, key: Any, value: Any, comment: Optional[Any] = ...
    ) -> None:
        """insert key value into given position, as defined by source YAML
        attach comment if provided
        """
        ...

    def mlget(self, key: Any, default: Any = ..., list_ok: Any = ...) -> Any:
        """multi-level get that expects dicts within dicts"""
        ...

    def __getitem__(self, key: Any) -> Any: ...
    def __setitem__(self, key: Any, value: Any) -> None: ...
    def __contains__(self, key: Any) -> bool: ...
    def get(self, key: Any, default: Any = ...) -> Any: ...
    def __repr__(self) -> Any: ...
    def non_merged_items(self) -> Any: ...
    def __delitem__(self, key: Any) -> None: ...
    def __iter__(self) -> Any: ...
    def pop(self, key: Any, default: Any = ...) -> Any: ...
    def __len__(self) -> int: ...
    def __eq__(self, other: Any) -> bool: ...
    def keys(self) -> Any: ...
    def values(self) -> Any: ...
    def items(self) -> Any: ...
    @property
    def merge(self) -> Any: ...
    def copy(self) -> Any: ...
    def add_referent(self, cm: Any) -> None: ...
    def add_yaml_merge(self, value: Any) -> None: ...
    def update_key_value(self, key: Any) -> None: ...
    def __deepcopy__(self, memo: Any) -> Any: ...

@classmethod
def raise_immutable(cls: Any, *args: Any, **kwargs: Any) -> None: ...

class CommentedKeyMap(CommentedBase, Mapping):
    __slots__ = ...
    def __init__(self, *args: Any, **kw: Any) -> None: ...

    update = ...
    def __getitem__(self, index: Any) -> Any: ...
    def __iter__(self) -> Iterator[Any]: ...
    def __len__(self) -> int: ...
    def __hash__(self) -> Any: ...
    def __repr__(self) -> Any: ...
    @classmethod
    def fromkeys(keys: Any, v: Any = ...) -> Any: ...

class CommentedOrderedMap(CommentedMap):
    __slots__ = ...

class CommentedSet(MutableSet, CommentedBase):
    __slots__ = ...
    def __init__(self, values: Any = ...) -> None: ...
    def add(self, value: Any) -> None:
        """Add an element."""
        ...

    def discard(self, value: Any) -> None:
        """Remove an element.  Do not raise an exception if absent."""
        ...

    def __contains__(self, x: Any) -> Any: ...
    def __iter__(self) -> Any: ...
    def __len__(self) -> int: ...
    def __repr__(self) -> str: ...

class TaggedScalar(CommentedBase):
    def __init__(self, value: Any = ..., style: Any = ..., tag: Any = ...) -> None: ...
    def __str__(self) -> Any: ...
    def count(
        self, s: str, start: Optional[int] = ..., end: Optional[int] = ...
    ) -> Any: ...
    def __getitem__(self, pos: int) -> Any: ...
    def __repr__(self) -> str: ...

def dump_comments(d: Any, name: str = ..., sep: str = ..., out: Any = ...) -> None:
    """
    recursively dump comments, all but the toplevel preceded by the path
    in dotted form x.0.a
    """
    ...



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypeAlias, TypedDict, cast

from snakebids.paths import specs
from snakebids.paths._factory import BidsFunction, bids_factory

if TYPE_CHECKING:
    from snakebids.paths._utils import BidsPathSpec

# <AUTOUPDATE>
# The code between these tags is automatically generated. Do not
# manually edit
# To update, run::
#
#       poetry run poe update-bids
#

VALID_SPECS: TypeAlias = Literal["v0_0_0", "v0_11_0", "v0_15_0", "latest"]
# </AUTOUPDATE>

__all__ = ["set_bids_spec"]


class _Config(TypedDict):
    active_spec: BidsPathSpec
    bids_func: BidsFunction


def set_bids_spec(spec: BidsPathSpec | VALID_SPECS):
    """Set the spec to be used by path generation functions (such as :func:`~snakebids.bids`).

    Parameters
    ----------
    spec
        Either a spec object, or the name of a builtin :ref:`spec <specs>`
    """  # noqa: E501
    if isinstance(spec, str):
        spec = cast("BidsPathSpec", getattr(specs, spec)())
    _config["active_spec"] = spec
    _config["bids_func"] = bids_factory(spec)


def get_bids_spec() -> BidsPathSpec:
    """Get the currently active BIDS spec."""
    return _config["active_spec"]


def get_bids_func() -> BidsFunction:
    """Get the currently active BIDS path build function."""
    return _config["bids_func"]


def reset_bids_spec():
    specs.__getattr__.cache_clear()  # type: ignore
    spec = specs.latest()
    _config["active_spec"] = spec
    _config["bids_func"] = bids_factory(spec, _implicit=True)


_config: _Config = {}  # type: ignore
reset_bids_spec()



from __future__ import annotations

import importlib.metadata as impm
from typing import Any, cast

import attrs

from snakebids import bidsapp
from snakebids.plugins.base import PluginBase


@attrs.define(kw_only=True)
class Version(PluginBase):
    """Expose app version in config.

    A version string can either be explicitly provided, or the name of the distribution.
    In the latter case, :mod:`importlib.metadata` will be used to look up the version.
    Note that this method does not work when the app is not installed; in this case, the
    version will be set to ``unknown``.

    Parameters
    ----------
    version
        Explicit version string
    distribution
        Name of distribution (as installed with ``pip``) for version lookup

    Raises
    ------
    ValueError
        If neither version nor distribution are provided, or if both are provided
    """

    version: str | None = None
    distribution: str | None = None

    PREFIX = "plugins.version"

    def __attrs_post_init__(self):
        if self.version is not None and self.distribution is not None:
            msg = "version and distribution may not both be specified"
            raise ValueError(msg)
        if self.version is None and self.distribution is None:
            msg = "One of version or distribution must be specified"
            raise ValueError(msg)

    @bidsapp.hookimpl
    def initialize_config(self, config: dict[str, Any]):
        """Assign version to config."""
        if self.version is not None:
            version = self.version
        else:
            try:
                version = impm.version(cast(str, self.distribution))
            except impm.PackageNotFoundError:
                version = "unknown"
        self.assign(config, "version", version)



from __future__ import annotations

import argparse
import subprocess as sp
from pathlib import Path

import pytest
from pytest_mock.plugin import MockerFixture

from snakebids.plugins.validator import BidsValidator, InvalidBidsError


def test_skip_validation():
    val = BidsValidator()
    parser = argparse.ArgumentParser()
    val.add_cli_arguments(parser)
    parsed = parser.parse_args(["--skip-bids-validation"])
    assert parsed.__dict__["plugins.validator.skip"] is True


class TestFinalizeConfig:
    def test_skip_validation(self):
        val = BidsValidator()
        # Test if validation is skipped
        config = {"plugins.validator.skip": True}

        val.finalize_config(config)
        assert "plugins.validator.success" not in config

    def test_validation_successful(self, mocker: MockerFixture):
        # Test successful validation
        mocker.patch("subprocess.check_call", return_value=0)
        val = BidsValidator()
        # Test if validation is skipped
        config = {"plugins.validator.skip": False, "bids_dir": Path()}

        val.finalize_config(config)

        assert config["plugins.validator.success"]

    def test_missing_bids_validator(
        self, mocker: MockerFixture, caplog: pytest.LogCaptureFixture
    ):
        # Test fall back to Pybids validation
        mocker.patch("subprocess.check_call", side_effect=FileNotFoundError)
        val = BidsValidator()
        # Test if validation is skipped
        config = {"plugins.validator.skip": False, "bids_dir": Path()}

        val.finalize_config(config)

        assert not config["plugins.validator.success"]

        # Check logger message (should only have 1 warning message)
        assert len(caplog.records) == 1
        assert caplog.records[0].levelname == "WARNING"
        assert "Missing bids-validator installation" in caplog.records[0].message

    def test_raise_validation_error(self, mocker: MockerFixture):
        # Test for any other bids-validation error
        mocker.patch("subprocess.check_call", side_effect=sp.CalledProcessError(1, ""))
        val = BidsValidator()
        # Test if validation is skipped
        config = {"plugins.validator.skip": False, "bids_dir": Path()}

        with pytest.raises(InvalidBidsError):
            val.finalize_config(config)

        # Check validation failure
        assert not config["plugins.validator.success"]

    def test_ignore_validation_error(self, mocker: MockerFixture):
        # Test for any other bids-validation error
        mocker.patch("subprocess.check_call", side_effect=sp.CalledProcessError(1, ""))
        val = BidsValidator(raise_invalid_bids=False)
        # Test if validation is skipped
        config = {"plugins.validator.skip": False, "bids_dir": Path()}
        val.finalize_config(config)

        # Check validation failure
        assert not config["plugins.validator.success"]



from .args import (
    ArgumentGroups,
    SnakebidsConfig,
)
from .hook import (
    hookimpl,
)
from .run import (
    ArgumentParserArgs,
    app,
)

__all__ = ["ArgumentGroups", "ArgumentParserArgs", "SnakebidsConfig", "app", "hookimpl"]



"""
This type stub file was generated by pyright.
"""

if False: ...
SHOW_LINES = ...

def CommentCheck() -> None: ...

class Event:
    __slots__ = ...
    crepr = ...
    def __init__(
        self, start_mark: Any = ..., end_mark: Any = ..., comment: Any = ...
    ) -> None: ...
    def __repr__(self) -> Any: ...
    def compact_repr(self) -> str: ...

class NodeEvent(Event):
    __slots__ = ...
    def __init__(
        self,
        anchor: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        comment: Any = ...,
    ) -> None: ...

class CollectionStartEvent(NodeEvent):
    __slots__ = ...
    def __init__(
        self,
        anchor: Any,
        tag: Any,
        implicit: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        flow_style: Any = ...,
        comment: Any = ...,
        nr_items: Optional[int] = ...,
    ) -> None: ...
    @property
    def tag(self) -> Optional[str]: ...

class CollectionEndEvent(Event):
    __slots__ = ...

class StreamStartEvent(Event):
    __slots__ = ...
    crepr = ...
    def __init__(
        self,
        start_mark: Any = ...,
        end_mark: Any = ...,
        encoding: Any = ...,
        comment: Any = ...,
    ) -> None: ...

class StreamEndEvent(Event):
    __slots__ = ...
    crepr = ...

class DocumentStartEvent(Event):
    __slots__ = ...
    crepr = ...
    def __init__(
        self,
        start_mark: Any = ...,
        end_mark: Any = ...,
        explicit: Any = ...,
        version: Any = ...,
        tags: Any = ...,
        comment: Any = ...,
    ) -> None: ...
    def compact_repr(self) -> str: ...

class DocumentEndEvent(Event):
    __slots__ = ...
    crepr = ...
    def __init__(
        self,
        start_mark: Any = ...,
        end_mark: Any = ...,
        explicit: Any = ...,
        comment: Any = ...,
    ) -> None: ...
    def compact_repr(self) -> str: ...

class AliasEvent(NodeEvent):
    __slots__ = ...
    crepr = ...
    def __init__(
        self,
        anchor: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        style: Any = ...,
        comment: Any = ...,
    ) -> None: ...
    def compact_repr(self) -> str: ...

class ScalarEvent(NodeEvent):
    __slots__ = ...
    crepr = ...
    def __init__(
        self,
        anchor: Any,
        tag: Any,
        implicit: Any,
        value: Any,
        start_mark: Any = ...,
        end_mark: Any = ...,
        style: Any = ...,
        comment: Any = ...,
    ) -> None: ...
    @property
    def tag(self) -> Optional[str]: ...
    @tag.setter
    def tag(self, val: Any) -> None: ...
    def compact_repr(self) -> str: ...

class SequenceStartEvent(CollectionStartEvent):
    __slots__ = ...
    crepr = ...
    def compact_repr(self) -> str: ...

class SequenceEndEvent(CollectionEndEvent):
    __slots__ = ...
    crepr = ...

class MappingStartEvent(CollectionStartEvent):
    __slots__ = ...
    crepr = ...
    def compact_repr(self) -> str: ...

class MappingEndEvent(CollectionEndEvent):
    __slots__ = ...
    crepr = ...



from __future__ import annotations

import copy
import itertools as it
import re
import string
import warnings
from pathlib import Path
from typing import Any

import more_itertools as itx
import pytest
from hypothesis import assume, given
from hypothesis import strategies as st

from snakebids.core.datasets import (
    BidsComponent,
    BidsComponentRow,
    BidsDataset,
    BidsPartialComponent,
)
from snakebids.exceptions import DuplicateComponentError
from snakebids.paths._presets import bids
from snakebids.snakemake_compat import WildcardError
from snakebids.types import Expandable, ZipList
from snakebids.utils import sb_itertools as sb_it
from snakebids.utils.snakemake_io import glob_wildcards
from snakebids.utils.utils import BidsEntity, get_wildcard_dict, zip_list_eq
from tests import strategies as sb_st
from tests.helpers import (
    Benchmark,
    expand_zip_list,
    get_bids_path,
    get_zip_list,
    setify,
)


def test_multiple_components_cannot_have_same_name():
    comp1 = BidsComponent(name="foo", path=".", zip_lists={})
    comp2 = BidsComponent(name="foo", path=".", zip_lists={})
    with pytest.raises(DuplicateComponentError):
        BidsDataset.from_iterable([comp1, comp2])


class TestBidsComponentAliases:
    @given(sb_st.bids_components())
    def test_bids_component_aliases_are_correctly_set(self, component: BidsComponent):
        assert component.input_path is component.path
        assert component.input_zip_lists is component.zip_lists
        assert component.input_lists is component.input_lists
        assert component.input_wildcards is component.wildcards

    @given(sb_st.bids_components())
    def test_bids_dataset_aliases_are_correctly_set(self, component: BidsComponent):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            dataset = BidsDataset.from_iterable([component])
            assert dataset.input_path == dataset.path
            assert dataset.input_zip_lists == dataset.zip_lists
            assert dataset.input_lists == dataset.entities
            assert dataset.input_wildcards == dataset.wildcards


class TestBidsComponentValidation:
    @given(sb_st.zip_lists().filter(lambda v: len(v) > 1))
    def test_zip_lists_must_be_same_length(self, zip_lists: ZipList):
        itx.first(zip_lists.values()).append("foo")
        with pytest.raises(ValueError, match="zip_lists must all be of equal length"):
            BidsComponent(
                name="foo", path=get_bids_path(zip_lists), zip_lists=zip_lists
            )

    @given(sb_st.zip_lists(), sb_st.bids_entity())
    def test_path_cannot_have_extra_entities(
        self, zip_lists: ZipList, entity: BidsEntity
    ):
        assume(entity.wildcard not in zip_lists)
        path = get_bids_path(it.chain(zip_lists, [entity.entity]))
        with pytest.raises(
            ValueError, match="zip_lists entries must match the wildcards in input_path"
        ):
            BidsComponent(name="foo", path=path, zip_lists=zip_lists)

    @given(sb_st.zip_lists().filter(lambda v: len(v) > 1))
    def test_path_cannot_have_missing_entities(self, zip_lists: ZipList):
        # Snakebids strategies won't return a zip_list with just datatype, but now that
        # we've dropped an entity we need to check again
        path_entities = sb_it.drop(1, zip_lists)
        assume(set(path_entities) - {"datatype"})

        path = get_bids_path(path_entities)
        with pytest.raises(
            ValueError, match="zip_lists entries must match the wildcards in input_path"
        ):
            BidsComponent(name="foo", path=path, zip_lists=zip_lists)


class TestBidsComponentEq:
    @given(sb_st.bids_components(), sb_st.everything_except(BidsComponent))
    def test_other_types_are_unequal(self, comp: BidsComponent, other: Any):
        assert comp != other

    def test_empty_bidsinput_are_equal(self):
        assert BidsComponent(name="", path="", zip_lists={}) == BidsComponent(
            name="", path="", zip_lists={}
        )
        assert BidsComponent(
            name="",
            path="{foo}{bar}",
            zip_lists={"foo": [], "bar": []},
        ) == BidsComponent(
            name="",
            path="{foo}{bar}",
            zip_lists={"foo": [], "bar": []},
        )

    @given(sb_st.bids_components())
    def test_copies_are_equal(self, comp: BidsComponent):
        cp = copy.deepcopy(comp)
        assert cp == comp

    @given(sb_st.bids_components())
    def test_mutation_makes_unequal(self, comp: BidsComponent):
        cp = copy.deepcopy(comp)
        itx.first(cp.zip_lists.values())[0] += "foo"
        assert cp != comp

    @given(sb_st.bids_components(), st.data())
    def test_extra_entities_makes_unequal(
        self, comp: BidsComponent, data: st.DataObject
    ):
        cp = copy.deepcopy(comp)
        new_entity = data.draw(
            sb_st.bids_value().filter(lambda s: s not in comp.zip_lists)
        )
        cp.zip_lists[new_entity] = []
        itx.first(cp.zip_lists.values())[0] += "foo"
        assert cp != comp

    @given(sb_st.bids_components())
    def test_order_doesnt_affect_equality(self, comp: BidsComponent):
        cp = copy.deepcopy(comp)
        for list_ in cp.zip_lists:
            cp.zip_lists[list_].reverse()
        assert cp == comp

    @given(sb_st.bids_components())
    def test_paths_must_be_identical(self, comp: BidsComponent):
        cp = BidsComponent(
            name=comp.input_name,
            path=comp.input_path + "foo",
            zip_lists=comp.zip_lists,
        )
        assert cp != comp


class TestBidsComponentProperties:
    @given(st.data(), st.integers(min_value=1, max_value=2))
    def test_input_lists_derives_from_zip_lists(
        self, data: st.DataObject, min_size: int
    ):
        input_lists = data.draw(sb_st.bids_input_lists(min_size=min_size, max_size=5))

        # Due to the product, we can delete some of the combinations and still
        # regenerate our input_lists
        combs = list(it.product(*input_lists.values()))[min_size - 1 :]
        zip_lists = get_zip_list(input_lists, combs)
        path = get_bids_path(zip_lists)

        assert setify(
            BidsComponent(name="foo", path=path, zip_lists=zip_lists).entities
        ) == setify(input_lists)

    @given(
        st.dictionaries(
            sb_st.bids_entity().map(lambda e: e.wildcard),
            sb_st.bids_value("[^.]*"),
            min_size=1,
        ).filter(lambda v: list(v) != ["datatype"])
    )
    def test_input_wildcards_derives_from_zip_lists(
        self,
        bids_entities: dict[str, str],
    ):
        zip_lists = {entity: [val] for entity, val in bids_entities.items()}
        bids_input = BidsComponent(
            name="foo",
            path=get_bids_path(zip_lists),
            zip_lists=zip_lists,
        )

        wildstr = ".".join(bids_input.input_wildcards.values())
        first = wildstr.format(**bids_input.input_wildcards)
        second = first.format(**bids_entities)
        assert set(second.split(".")) == set(bids_entities.values())


legacy_names = [
    "input_path",
    "input_zip_lists",
    "input_lists",
    "input_wildcards",
    "subjects",
    "sessions",
    "subj_wildcards",
]


class TestBidsDatasetLegacyAccess:
    match_str = r".*generate_inputs\(\) no longer returns a dict by default.*"

    @given(dataset=sb_st.datasets(), name=st.sampled_from(legacy_names))
    def test_accessing_legacy_names_leads_to_informative_error(
        self, dataset: BidsDataset, name: str
    ):
        assume(name not in dataset)
        with pytest.raises(KeyError, match=self.match_str):
            dataset[name]

    @given(
        dataset=sb_st.datasets(),
        name=st.text(sb_st.alphanum).filter(lambda s: s not in legacy_names),
    )
    def test_accessing_ordinary_missing_names_leads_to_regular_error(
        self, dataset: BidsDataset, name: str
    ):
        assume(name not in dataset)
        with pytest.raises(KeyError) as err:
            dataset[name]

        assert not re.search(self.match_str, str(err.value))

    @given(
        dataset=sb_st.datasets_one_comp(
            names=st.shared(st.sampled_from(legacy_names), key="names")
        ),
        name=st.shared(st.sampled_from(legacy_names), key="names"),
    )
    def test_legacy_names_can_be_accessed_if_component_name(
        self, dataset: BidsDataset, name: str
    ):
        assert isinstance(dataset[name], BidsComponent)


def _get_novel_path(prefix: str, component: Expandable):
    # use the "comp-" prefix to give a constant part to the novel template,
    # otherwise the trivial template "{foo}" globs everything
    return Path(
        *(f"{prefix}-{s}" for s in get_wildcard_dict(component.zip_lists).values())
    )


def test_benchmark_zip_list_comparison(benchmark: Benchmark):
    iters = 1000
    first = {
        "nucleus": ["yit", "yit"] * iters,
        "subject": ["100", "200"] * iters,
        "mod": ["mUH", "Muh"] * iters,
        "flip": ["20755", "20755"] * iters,
    }
    second = {
        "subject": ["200", "100"] * iters,
        "nucleus": ["yit", "yit"] * iters,
        "flip": ["20755", "20755"] * iters,
        "mod": ["Muh", "mUH"] * iters,
    }
    assert benchmark(zip_list_eq, first, second)


class TestZipListEq:
    @given(first=sb_st.zip_lists(max_entities=5), data=st.data())
    def test_zip_list_equality_with_different_key_order(
        self, first: ZipList, data: st.DataObject
    ):
        order = data.draw(st.permutations(list(first)))
        second = {key: first[key] for key in order}
        assert zip_list_eq(first, second)

    @given(first=sb_st.zip_lists(), data=st.data())
    def test_zip_list_equality_with_different_value_order(
        self, first: ZipList, data: st.DataObject
    ):
        order = data.draw(st.permutations(range(len(itx.first(first.values())))))
        second = {key: [first[key][i] for i in order] for key in first}
        assert zip_list_eq(first, second)

    def test_zip_list_inequality(self):
        # Developing a generic strategy for inequal ziplists is actually nontrivial
        # Here at least we test with a specific example to prove the principle
        first = {
            "subject": [str(val) for val in range(5)],
            "session": [str(val) for val in range(5)],
        }
        second = copy.deepcopy(first)
        second["subject"] = list(reversed(second["subject"]))
        assert not zip_list_eq(first, second)


class TestExpandables:
    @given(
        component=sb_st.expandables(
            restrict_patterns=True, unique=True, blacklist_entities=["extension"]
        ),
        wildcards=st.lists(
            st.text(string.ascii_letters, min_size=1, max_size=10).filter(
                lambda s: s not in sb_st.valid_entities
            ),
            min_size=1,
            max_size=5,
        ),
        data=st.data(),
    )
    def test_expand_with_extra_args_returns_all_paths(
        self, component: Expandable, wildcards: list[str], data: st.DataObject
    ):
        num_wildcards = len(wildcards)
        values = data.draw(
            st.lists(
                st.lists(
                    st.text(sb_st.alphanum, min_size=1, max_size=10),
                    min_size=1,
                    max_size=3,
                ),
                min_size=num_wildcards,
                max_size=num_wildcards,
            )
        )
        path_tpl = bids(
            **get_wildcard_dict(component.zip_lists),
            **get_wildcard_dict(wildcards),
        )
        wcard_dict = dict(zip(wildcards, values, strict=True))
        zlist = expand_zip_list(component.zip_lists, wcard_dict)
        paths = component.expand(path_tpl, **wcard_dict)
        assert zip_list_eq(glob_wildcards(path_tpl, paths), zlist)

    @given(
        component=sb_st.expandables(
            restrict_patterns=True, blacklist_entities=["extension"]
        )
    )
    def test_expand_over_multiple_paths(self, component: Expandable):
        path1 = _get_novel_path("first", component)
        path2 = _get_novel_path("second", component)
        paths = component.expand([path1, path2])
        assert zip_list_eq(glob_wildcards(path1, paths), glob_wildcards(path2, paths))

    @given(
        component=sb_st.expandables(restrict_patterns=True),
        wildcard=st.text(string.ascii_letters, min_size=1, max_size=10).filter(
            lambda s: s not in sb_st.valid_entities
        ),
    )
    def test_partial_expansion(self, component: Expandable, wildcard: str):
        path_tpl = bids(
            **get_wildcard_dict(component.zip_lists), **get_wildcard_dict(wildcard)
        )
        paths = component.expand(path_tpl, allow_missing=True)
        for path in paths:
            assert re.search(r"\{.+\}", path)

    @given(
        component=sb_st.expandables(restrict_patterns=True),
        wildcard=st.text(string.ascii_letters, min_size=1, max_size=10).filter(
            lambda s: s not in sb_st.valid_entities
        ),
    )
    def test_prevent_partial_expansion(self, component: Expandable, wildcard: str):
        path_tpl = bids(
            **get_wildcard_dict(component.zip_lists), **get_wildcard_dict(wildcard)
        )
        with pytest.raises(WildcardError):
            component.expand(path_tpl)

    @given(component=sb_st.expandables(restrict_patterns=True, path_safe=True))
    def test_expand_deduplicates_paths(self, component: Expandable):
        path_tpl = bids(**get_wildcard_dict(component.zip_lists))
        paths = component.expand(path_tpl)
        assert len(paths) == len(set(paths))

    @given(
        component=sb_st.expandables(
            restrict_patterns=True,
            path_safe=True,
            unique=True,
        )
    )
    def test_expand_preserves_entry_order(self, component: Expandable):
        path_tpl = bids(**get_wildcard_dict(component.zip_lists))
        paths = component.expand(path_tpl)
        for path, entity_vals in zip(
            paths, zip(*component.zip_lists.values(), strict=True), strict=True
        ):
            assert (
                bids(**dict(zip(component.zip_lists.keys(), entity_vals, strict=True)))
                == path
            )

    @given(path=st.text())
    def test_expandable_with_no_wildcards_returns_path_unaltered(self, path: str):
        component = BidsPartialComponent(zip_lists={})
        assert itx.one(component.expand(path)) == path

    @given(component=sb_st.expandables(min_values=0, max_values=0, path_safe=True))
    def test_expandable_with_no_entries_returns_empty_list(self, component: Expandable):
        path_tpl = bids(**get_wildcard_dict(component.zip_lists))
        assert component.expand(path_tpl) == []


class TestBidsComponentExpand:
    """
    `extension` is generally excluded from the generated components because its wildcard
    is immediately adjacent to the previous wildcard (e.g.
    sub-{subject}_{suffix}{extension}), making it impossible to correctly parse using
    `glob_wildcards`. Its absence does not affect the spirit of the tests.

    We also exclude extra_entities from the generated Components for the same reason:
    ``glob_wildcards`` would glob the extra entities found in the generated path, making
    comparison impossible
    """

    @given(
        component=sb_st.bids_components(
            blacklist_entities=["extension"],
            restrict_patterns=True,
            extra_entities=False,
            unique=True,
        )
    )
    def test_expand_with_no_args_returns_initial_paths(self, component: BidsComponent):
        paths = component.expand()
        assert zip_list_eq(glob_wildcards(component.path, paths), component.zip_lists)

    @given(
        component=sb_st.bids_components(restrict_patterns=True, extra_entities=False)
    )
    def test_not_expand_over_internal_path_when_novel_given(
        self, component: BidsComponent
    ):
        assume(set(component.wildcards) - {"datatype", "suffix", "extension"})
        novel_path = _get_novel_path("comp", component)
        paths = component.expand(novel_path)
        assert not glob_wildcards(component.path, paths)


class TestFiltering:
    def get_filter_dict(
        self,
        data: st.DataObject,
        component: Expandable,
        allow_extra_filters: bool = False,
    ):
        filter_dict: dict[str, list[str] | str] = {}
        filters = data.draw(
            st.lists(
                st.one_of(
                    [
                        st.sampled_from(list(component.zip_lists)),
                        st.text(string.ascii_letters, min_size=1, max_size=5),
                    ]
                )
                if allow_extra_filters
                else st.sampled_from(list(component.zip_lists)),
                max_size=len(component.zip_lists),
                unique=True,
            )
        )

        entities = {key: list(set(vals)) for key, vals in component.zip_lists.items()}

        def value_strat(filt: str):
            rand_text = st.text(sb_st.alphanum, min_size=1, max_size=10)
            return (
                st.one_of([st.sampled_from(entities[filt]), rand_text])
                if filt in entities
                else rand_text
            )

        for filt in filters:
            filter_dict[filt] = data.draw(
                st.one_of(
                    [
                        st.lists(
                            value_strat(filt),
                            unique=True,
                            min_size=1,
                            max_size=5,
                        ),
                        value_strat(filt),
                    ]
                )
            )
        return filter_dict

    def filter_iter(self, filters: dict[str, str | list[str]]):
        return {
            key: iter(val) if isinstance(val, list) else val
            for key, val in filters.items()
        }

    @given(
        component=sb_st.bids_components(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_only_filter_values_in_output(
        self, component: Expandable, data: st.DataObject
    ):
        filter_dict = self.get_filter_dict(data, component)
        filtered = component.filter(**self.filter_iter(filter_dict))
        for filt in filter_dict:
            for val in filtered.zip_lists[filt]:
                assert val in filter_dict[filt]

    @given(
        component=sb_st.expandables(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_zip_lists_rows_remain_of_equal_length(
        self, component: Expandable, data: st.DataObject
    ):
        filter_dict = self.get_filter_dict(data, component)
        filtered = component.filter(**self.filter_iter(filter_dict))
        lengths: set[int] = set()
        for row in filtered.zip_lists.values():
            lengths.add(len(row))
        assert len(lengths) == 1

    @given(
        component=sb_st.expandables(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_all_columns_found_in_original_zip_list(
        self, component: Expandable, data: st.DataObject
    ):
        filter_dict = self.get_filter_dict(data, component, allow_extra_filters=True)
        filtered = component.filter(**self.filter_iter(filter_dict))
        cols = set(zip(*component.zip_lists.values(), strict=True))
        for col in zip(*filtered.zip_lists.values(), strict=True):
            assert col in cols

    @given(
        component=sb_st.expandables(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_all_entities_remain_after_filtering(
        self, component: Expandable, data: st.DataObject
    ):
        filter_dict = self.get_filter_dict(data, component, allow_extra_filters=True)
        filtered = component.filter(**self.filter_iter(filter_dict))
        assert set(component.zip_lists) == set(filtered.zip_lists)

    @given(
        component=sb_st.expandables(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_no_columns_that_should_be_present_are_missing(
        self,
        component: Expandable,
        data: st.DataObject,
    ):
        filter_dict = self.get_filter_dict(data, component)
        filtered = component.filter(**self.filter_iter(filter_dict))
        keys = list(component.zip_lists)
        cols = set(zip(*component.zip_lists.values(), strict=True))
        result = set(zip(*filtered.zip_lists.values(), strict=True))
        for col in cols:
            should_be_present = True
            for filt in filter_dict:
                if col[keys.index(filt)] not in list(
                    itx.always_iterable(filter_dict[filt])
                ):
                    assert col not in result
                    should_be_present = False
                    break
            if should_be_present:
                assert col in result


class TestFilteringBidsComponentRowWithSpec:
    def get_filter_spec(
        self,
        data: st.DataObject,
        component: BidsComponentRow,
    ):
        rand_text = st.text(sb_st.alphanum, min_size=1, max_size=10)
        value_strat = st.one_of(st.sampled_from(component.entities), rand_text)

        return data.draw(
            st.one_of(
                st.lists(
                    value_strat,
                    unique=True,
                    min_size=1,
                    max_size=5,
                ),
                value_strat,
            )
        )

    @given(
        component=sb_st.bids_component_row(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_only_filter_values_in_output(
        self, component: BidsComponentRow, data: st.DataObject
    ):
        spec = self.get_filter_spec(data, component)
        filtered = component.filter(spec)
        for val in filtered.entities:
            assert val in list(itx.always_iterable(spec))

    @given(
        component=sb_st.bids_component_row(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_all_columns_found_in_original_zip_list(
        self, component: BidsComponentRow, data: st.DataObject
    ):
        spec = self.get_filter_spec(data, component)
        filtered = component.filter(spec)
        cols = set(zip(*component.zip_lists.values(), strict=True))
        for col in zip(*filtered.zip_lists.values(), strict=True):
            assert col in cols

    @given(
        component=sb_st.bids_component_row(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_all_entities_remain_after_filtering(
        self, component: BidsComponentRow, data: st.DataObject
    ):
        spec = self.get_filter_spec(data, component)
        filtered = component.filter(spec)
        assert set(component.zip_lists) == set(filtered.zip_lists)

    @given(
        component=sb_st.bids_component_row(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_all_valid_values_in_spec_in_result(
        self, component: BidsComponentRow, data: st.DataObject
    ):
        spec = self.get_filter_spec(data, component)
        filtered = component.filter(itx.always_iterable(spec))
        for val in itx.always_iterable(spec):
            if val in component.entities:
                assert val in filtered.entities

    @given(
        component=sb_st.bids_component_row(max_values=4, restrict_patterns=True),
        data=st.data(),
    )
    def test_providing_both_spec_and_filters_gives_error(
        self, component: BidsComponentRow, data: st.DataObject
    ):
        spec = self.get_filter_spec(data, component)
        with pytest.raises(ValueError, match="__spec and filters cannot"):
            component.filter(spec, foo="bar")


class TestBidsComponentIndexing:
    def get_selectors(
        self,
        data: st.DataObject,
        dicts: BidsPartialComponent,
        use_nonexistant_keys: bool = False,
        unique: bool = False,
    ) -> tuple[str, ...]:
        if not dicts.zip_lists and not use_nonexistant_keys:
            return ()
        sampler = st.sampled_from(list(dicts.zip_lists))
        val_strat = (
            st.text().filter(lambda s: s not in dicts.zip_lists)
            if use_nonexistant_keys
            else sampler
        )
        return tuple(data.draw(st.lists(val_strat, unique=unique)))

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_multiple_selection_returns_bidspartialcomponent(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        assert isinstance(dicts[selectors], BidsPartialComponent)

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_single_selection_returns_bidscomponentrow(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = data.draw(st.sampled_from(list(dicts.zip_lists)))
        assert isinstance(dicts[selectors], BidsComponentRow)

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_selected_items_in_original(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        selected = dicts[selectors]
        for key in selected.zip_lists:
            assert selected[key] == dicts[key]

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_all_requested_items_received_and_no_others(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        selected = dicts[selectors]
        assert set(selectors) == set(selected.zip_lists)
        for selector in selectors:
            assert selector in selected.zip_lists

    @given(
        dicts=sb_st.bids_partial_components(),
        data=st.data(),
    )
    def test_single_missing_key_raises_error(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selector = data.draw(st.text().filter(lambda s: s not in dicts.zip_lists))
        with pytest.raises(KeyError):
            dicts[selector]

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_multiple_missing_key_raises_error(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts, use_nonexistant_keys=True)
        assume(len(selectors))
        with pytest.raises(KeyError):
            dicts[selectors]

    @given(dicts=sb_st.bids_partial_components(), data=st.data())
    def test_order_of_selectors_is_preserved(
        self, dicts: BidsPartialComponent, data: st.DataObject
    ):
        selectors = self.get_selectors(data, dicts)
        # Using the itx method for uniqueness to avoid calculating unique values in the
        # test in the same way as the source code
        assert tuple(dicts.zip_lists[selectors]) == tuple(
            itx.unique_everseen(selectors)
        )



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.error import MarkedYAMLError

if False: ...
__all__ = ["Composer", "ComposerError"]

class ComposerError(MarkedYAMLError): ...

class Composer:
    def __init__(self, loader: Any = ...) -> None: ...
    @property
    def parser(self) -> Any: ...
    @property
    def resolver(self) -> Any: ...
    def check_node(self) -> Any: ...
    def get_node(self) -> Any: ...
    def get_single_node(self) -> Any: ...
    def compose_document(self: Any) -> Any: ...
    def return_alias(self, a: Any) -> Any: ...
    def compose_node(self, parent: Any, index: Any) -> Any: ...
    def compose_scalar_node(self, anchor: Any) -> Any: ...
    def compose_sequence_node(self, anchor: Any) -> Any: ...
    def compose_mapping_node(self, anchor: Any) -> Any: ...
    def check_end_doc_comment(self, end_event: Any, node: Any) -> None: ...



from __future__ import annotations

import argparse
import logging
from pathlib import Path
from typing import Any

import pytest
from hypothesis import given
from hypothesis import strategies as st

from snakebids.plugins.pybidsdb import Pybidsdb
from snakebids.utils.utils import DEPRECATION_FLAG
from tests import strategies as sb_st


class TestAddArguments:
    def test_args_not_added_if_already_present(self):
        pybidsdb = Pybidsdb()
        parser = argparse.ArgumentParser()
        parser.add_argument("--pybidsdb-dir", dest="plugins.pybidsdb.dir")
        parser.add_argument("--pybidsdb-reset", dest="plugins.pybidsdb.reset")
        parser.add_argument("--reset-db", dest="plugins.pybidsdb.reset_db")
        pybidsdb.add_cli_arguments(parser, {})
        args = parser.parse_args(
            ["--pybidsdb-dir", "foo", "--pybidsdb-reset", "foo", "--reset-db", "foo"]
        )
        args.__dict__ = {
            "pybidsdb_dir": "foo",
            "pybidsdb_reset": "foo",
            "reset_db": "foo",
        }

    @given(title=st.text() | st.none())
    def test_args_put_in_specified_group(self, title: str | None):
        pybidsdb = Pybidsdb(title)
        parser = argparse.ArgumentParser()
        groups = {title: parser.add_argument_group(title)} if title is not None else {}
        pybidsdb.add_cli_arguments(parser, groups)

        if title is not None:
            assert groups[title]._actions
        assert parser._actions


class TestUpdateNamespace:
    @given(path=sb_st.paths())
    def test_pybidsdb_path_resolved(self, path: Path):
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": path,
            f"{pybidsdb.PREFIX}.reset": False,
            f"{pybidsdb.PREFIX}.reset_db": False,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)
        # Prepare app and initial config values
        assert config["pybidsdb_dir"] == path.resolve()

    @pytest.mark.parametrize("value", [True, False])
    def test_pybidsdb_reset_and_reset_db_both_work(self, value: bool):
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": Path(),
            f"{pybidsdb.PREFIX}.reset": value,
            f"{pybidsdb.PREFIX}.reset_db": not value,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)
        # Prepare app and initial config values
        assert config["pybidsdb_reset"] is True

    def test_warning_issued_for_reset_db(self, caplog: pytest.LogCaptureFixture):
        caplog.set_level(logging.WARNING)
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": Path(),
            f"{pybidsdb.PREFIX}.reset": True,
            f"{pybidsdb.PREFIX}.reset_db": True,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)
        # Prepare app and initial config values
        assert "--reset-db/--reset_db will be deprecated" in caplog.text

    @pytest.mark.parametrize("value", [True, False])
    def test_namespace_cleared(self, value: bool):
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": Path(),
            f"{pybidsdb.PREFIX}.reset": value,
            f"{pybidsdb.PREFIX}.reset_db": not value,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)
        # Prepare app and initial config values
        assert namespace == {}

    @pytest.mark.parametrize("value", [True, False])
    @given(path=sb_st.paths())
    def test_old_config_keys_deprecated(self, path: Path, value: bool):
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": path,
            f"{pybidsdb.PREFIX}.reset": value,
            f"{pybidsdb.PREFIX}.reset_db": False,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)

        assert (
            config["pybids_db_dir"]
            == f"{DEPRECATION_FLAG}{path.resolve()}{DEPRECATION_FLAG}"
        )
        assert (
            config["pybids_db_reset"]
            == f"{DEPRECATION_FLAG}{int(value)}{DEPRECATION_FLAG}"
        )

    def test_old_key_not_set_if_pybidsdb_not_set(self):
        pybidsdb = Pybidsdb()

        namespace = {
            f"{pybidsdb.PREFIX}.dir": None,
            f"{pybidsdb.PREFIX}.reset": False,
            f"{pybidsdb.PREFIX}.reset_db": False,
        }
        config: dict[str, Any] = {}
        pybidsdb.update_cli_namespace(namespace, config)

        assert config["pybids_db_dir"] is None



import sys
from collections.abc import Callable
from enum import Enum
from typing import Any, AnyStr, NoReturn

from pyfakefs.fake_file import AnyFile, AnyFileWrapper, FakeDirectory, FakeFile
from pyfakefs.helpers import AnyPath, AnyString

if sys.platform.startswith("linux"):
    _MAX_LINK_DEPTH = ...
else:
    _MAX_LINK_DEPTH = ...

class OSType(Enum):
    LINUX = ...
    MACOS = ...
    WINDOWS = ...

FakeNullFile = ...
FakeFileFromRealFile = ...
FakeDirectoryFromRealDirectory = ...
FakeFileWrapper = ...
StandardStreamWrapper = ...
FakeDirWrapper = ...
FakePipeWrapper = ...
FakePathModule = ...
FakeOsModule = ...
FakeFileOpen = ...
FakeIoModule = ...
if sys.platform != "win32":
    FakeFcntlModule = ...
PatchMode = ...
is_root = ...
set_uid = ...
set_gid = ...
reset_ids = ...

class FakeFilesystem:
    def __init__(
        self,
        path_separator: str = ...,
        total_size: int | None = ...,
        patcher: Any = ...,
    ) -> None: ...
    @property
    def is_linux(self) -> bool: ...
    @property
    def is_windows_fs(self) -> bool: ...
    @is_windows_fs.setter
    def is_windows_fs(self, value: bool) -> None: ...
    @property
    def is_macos(self) -> bool: ...
    @is_macos.setter
    def is_macos(self, value: bool) -> None: ...
    @property
    def cwd(self) -> str: ...
    @cwd.setter
    def cwd(self, value: str) -> None: ...
    @property
    def root_dir(self) -> FakeDirectory: ...
    @property
    def root_dir_name(self) -> str: ...
    @property
    def os(self) -> OSType: ...
    @os.setter
    def os(self, value: OSType) -> None: ...
    def reset(self, total_size: int | None = ...) -> None: ...
    def pause(self) -> None: ...
    def resume(self) -> None: ...
    def clear_cache(self) -> None: ...
    def line_separator(self) -> str: ...
    def raise_os_error(
        self,
        err_no: int,
        filename: AnyString | None = ...,
        winerror: int | None = ...,
    ) -> NoReturn: ...
    def get_path_separator(self, path: AnyStr) -> AnyStr: ...
    def starts_with_sep(self, path: AnyStr) -> bool: ...
    def add_mount_point(
        self, path: AnyStr, total_size: int | None = ..., can_exist: bool = ...
    ) -> dict: ...
    def get_disk_usage(self, path: AnyStr | None = ...) -> tuple[int, int, int]: ...
    def set_disk_usage(self, total_size: int, path: AnyStr | None = ...) -> None: ...
    def change_disk_usage(
        self, usage_change: int, file_path: AnyStr, st_dev: int
    ) -> None: ...
    def stat(self, entry_path: AnyStr, follow_symlinks: bool = ...): ...
    def raise_for_filepath_ending_with_separator(
        self,
        entry_path: AnyStr,
        file_object: FakeFile,
        follow_symlinks: bool = ...,
        macos_handling: bool = ...,
    ) -> None: ...
    def chmod(
        self,
        path: AnyStr,
        mode: int,
        follow_symlinks: bool = ...,
        force_unix_mode: bool = ...,
    ) -> None: ...
    def utime(
        self,
        path: AnyStr,
        times: tuple[int | float, int | float] | None = ...,
        *,
        ns: tuple[int, int] | None = ...,
        follow_symlinks: bool = ...,
    ) -> None: ...
    def get_open_file(self, file_des: int) -> AnyFileWrapper: ...
    def has_open_file(self, file_object: FakeFile) -> bool: ...
    def normcase(self, path: AnyStr) -> AnyStr: ...
    def normpath(self, path: AnyStr) -> AnyStr: ...
    def absnormpath(self, path: AnyStr) -> AnyStr: ...
    def splitpath(self, path: AnyStr) -> tuple[AnyStr, AnyStr]: ...
    def splitdrive(self, path: AnyStr) -> tuple[AnyStr, AnyStr]: ...
    def splitroot(self, path: AnyStr): ...
    def joinpaths(self, *paths: AnyStr) -> AnyStr: ...
    def starts_with_drive_letter(self, file_path: AnyStr) -> bool: ...
    def replace_windows_root(self, path: AnyStr) -> AnyStr: ...
    def is_mount_point(self, file_path: AnyStr) -> bool: ...
    def ends_with_path_separator(self, path: int | AnyPath) -> bool: ...
    def is_filepath_ending_with_separator(self, path: AnyStr) -> bool: ...
    def exists(self, file_path: AnyPath, check_link: bool = ...) -> bool: ...
    def resolve_path(self, file_path: AnyStr, allow_fd: bool = ...) -> AnyStr: ...
    def get_object_from_normpath(
        self, file_path: AnyPath, check_read_perm: bool = ..., check_owner: bool = ...
    ) -> AnyFile: ...
    def get_object(
        self, file_path: AnyPath, check_read_perm: bool = ...
    ) -> FakeFile: ...
    def resolve(
        self,
        file_path: AnyStr,
        follow_symlinks: bool = ...,
        allow_fd: bool = ...,
        check_read_perm: bool = ...,
        check_owner: bool = ...,
    ) -> FakeFile: ...
    def lresolve(self, path: AnyPath) -> FakeFile: ...
    def add_object(self, file_path: AnyStr, file_object: AnyFile) -> None: ...
    def rename(
        self, old_file_path: AnyPath, new_file_path: AnyPath, force_replace: bool = ...
    ) -> None: ...
    def remove_object(self, file_path: AnyStr) -> None: ...
    def make_string_path(self, path: AnyPath) -> AnyStr: ...
    def create_dir(
        self, directory_path: AnyPath, perm_bits: int = ...
    ) -> FakeDirectory: ...
    def create_file(
        self,
        file_path: AnyPath[AnyStr],
        st_mode: int = ...,
        contents: AnyString = ...,
        st_size: int | None = ...,
        create_missing_dirs: bool = ...,
        apply_umask: bool = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        side_effect: Callable[[FakeFile], Any] | None = ...,
    ) -> FakeFile: ...
    def add_real_file(
        self,
        source_path: AnyPath[AnyStr],
        read_only: bool = ...,
        target_path: AnyPath[AnyStr] | None = ...,
    ) -> FakeFile: ...
    def add_real_symlink(
        self, source_path: AnyPath, target_path: AnyPath | None = ...
    ) -> FakeFile: ...
    def add_real_directory(
        self,
        source_path: AnyPath[AnyStr],
        read_only: bool = ...,
        lazy_read: bool = ...,
        target_path: AnyPath[AnyStr] | None = ...,
    ) -> FakeDirectory: ...
    def add_real_paths(
        self, path_list: list[AnyStr], read_only: bool = ..., lazy_dir_read: bool = ...
    ) -> None: ...
    def create_file_internally(
        self,
        file_path: AnyPath,
        st_mode: int = ...,
        contents: AnyString = ...,
        st_size: int | None = ...,
        create_missing_dirs: bool = ...,
        apply_umask: bool = ...,
        encoding: str | None = ...,
        errors: str | None = ...,
        read_from_real_fs: bool = ...,
        side_effect: Callable | None = ...,
    ) -> FakeFile: ...
    def create_symlink(
        self, file_path: AnyPath, link_target: AnyPath, create_missing_dirs: bool = ...
    ) -> FakeFile: ...
    def create_link(
        self,
        old_path: AnyPath,
        new_path: AnyPath,
        follow_symlinks: bool = ...,
        create_missing_dirs: bool = ...,
    ) -> FakeFile: ...
    def link(
        self, old_path: AnyPath, new_path: AnyPath, follow_symlinks: bool = ...
    ) -> FakeFile: ...
    def readlink(self, path: AnyPath) -> str: ...
    def makedir(self, dir_path: AnyPath, mode: int = ...) -> None: ...
    def makedirs(
        self, dir_name: AnyStr, mode: int = ..., exist_ok: bool = ...
    ) -> None: ...
    def isdir(self, path: AnyPath, follow_symlinks: bool = ...) -> bool: ...
    def isfile(self, path: AnyPath, follow_symlinks: bool = ...) -> bool: ...
    def islink(self, path: AnyPath) -> bool: ...
    def confirmdir(
        self, target_directory: AnyStr, check_owner: bool = ...
    ) -> FakeDirectory: ...
    def remove(self, path: AnyStr) -> None: ...
    def rmdir(self, target_directory: AnyStr, allow_symlink: bool = ...) -> None: ...
    def listdir(self, target_directory: AnyStr) -> list[AnyStr]: ...
    def __str__(self) -> str: ...



"""
This type stub file was generated by pyright.
"""

from bids_validator import BIDSValidator

from . import _version
from .due import Doi, due
from .layout import BIDSLayout, BIDSLayoutIndexer

__all__ = [
    "modeling",
    "BIDSLayout",
    "BIDSLayoutIndexer",
    "BIDSValidator",
    "config",
    "layout",
    "reports",
    "utils",
    "variables",
]
__version__ = _version.get_versions()["version"]



"""Utilities for converting Snakemake apps to BIDS apps."""

from __future__ import annotations

import json
import logging
import os
import re
import warnings
from collections import defaultdict
from collections.abc import Iterable
from pathlib import Path
from typing import (
    Any,
    Literal,
    overload,
)

import more_itertools as itx
from bids import BIDSLayout, BIDSLayoutIndexer

from snakebids.core._querying import (
    FilterSpecError,
    PostFilter,
    UnifiedFilter,
    get_matching_files,
)
from snakebids.core.datasets import BidsComponent, BidsDataset, BidsDatasetDict
from snakebids.core.filtering import filter_list
from snakebids.exceptions import (
    ConfigError,
    DuplicateComponentError,
    RunError,
)
from snakebids.snakemake_compat import Snakemake
from snakebids.types import InputConfig, InputsConfig, ZipList
from snakebids.utils.snakemake_io import glob_wildcards
from snakebids.utils.utils import (
    DEPRECATION_FLAG,
    BidsEntity,
    BidsParseError,
    get_first_dir,
)

_logger = logging.getLogger(__name__)


@overload
def generate_inputs(
    bids_dir: Path | str,
    pybids_inputs: InputsConfig,
    *,
    pybidsdb_dir: Path | str | None = ...,
    pybidsdb_reset: bool = ...,
    derivatives: bool | Path | str = ...,
    pybids_config: str | None = ...,
    limit_to: Iterable[str] | None = ...,
    participant_label: Iterable[str] | str | None = ...,
    exclude_participant_label: Iterable[str] | str | None = ...,
    use_bids_inputs: Literal[True] | None = ...,
    index_metadata: bool = ...,
    validate: bool = ...,
    pybids_database_dir: Path | str | None = ...,
    pybids_reset_database: bool = ...,
) -> BidsDataset: ...


@overload
def generate_inputs(
    bids_dir: Path | str,
    pybids_inputs: InputsConfig,
    *,
    pybidsdb_dir: Path | str | None = ...,
    pybidsdb_reset: bool = ...,
    derivatives: bool | Path | str = ...,
    pybids_config: str | None = ...,
    limit_to: Iterable[str] | None = ...,
    participant_label: Iterable[str] | str | None = ...,
    exclude_participant_label: Iterable[str] | str | None = ...,
    use_bids_inputs: Literal[False],
    index_metadata: bool = ...,
    validate: bool = ...,
    pybids_database_dir: Path | str | None = ...,
    pybids_reset_database: bool = ...,
) -> BidsDatasetDict: ...


def generate_inputs(
    bids_dir: Path | str,
    pybids_inputs: InputsConfig,
    *,
    pybidsdb_dir: Path | str | None = None,
    pybidsdb_reset: bool | None = None,
    derivatives: bool | Path | str = False,
    pybids_config: str | None = None,
    limit_to: Iterable[str] | None = None,
    participant_label: Iterable[str] | str | None = None,
    exclude_participant_label: Iterable[str] | str | None = None,
    use_bids_inputs: bool | None = None,
    index_metadata: bool = False,
    validate: bool = False,
    pybids_database_dir: Path | str | None = None,
    pybids_reset_database: bool | None = None,
) -> BidsDataset | BidsDatasetDict:
    """Dynamically generate snakemake inputs using pybids_inputs.

    Pybids is used to parse the bids_dir. Custom paths can also be parsed by including
    the custom_paths entry under the pybids_inputs descriptor.

    Parameters
    ----------
    bids_dir
        Path to bids directory

    pybids_inputs
        Configuration for bids inputs, with keys as the names (``str``)

        Nested `dicts` with the following required keys (for complete info, see
        :class:`~snakebids.types.InputConfig`):

        * ``"filters"``: Dictionary of entity: "values" (dict of str -> str or list of
          str). The entity keywords should the bids tags on which to filter. The values
          should be an acceptable str value for that entity, or a list of acceptable str
          values.

        * ``"wildcards"``: List of (str) bids tags to include as wildcards in
          snakemake. At minimum this should usually include
          ``['subject','session']``, plus any other wildcards that you may
          want to make use of in your snakemake workflow, or want to retain
          in the output paths. Any wildcards in this list that are not in the
          filename will just be ignored.

        * ``"custom_path"``: Custom path to be parsed with wildcards wrapped in braces,
          as in ``/path/to/sub-{subject}/{wildcard_1}-{wildcard_2}``. This path will be
          parsed without pybids, allowing the use of non-bids-compliant paths.

    pybidsdb_dir
        Path to database directory. If None is provided, database
        is not used

    pybidsdb_reset
        A boolean that determines whether to reset / overwrite
        existing database.

    derivatives
        Indicates whether pybids should look for derivative datasets under bids_dir.
        These datasets must be properly formatted according to bids specs to be
        recognized. Defaults to False.

    limit_to
        If provided, indicates which input descriptors from pybids_inputs should be
        parsed. For example, if pybids_inputs describes ``"bold"`` and ``"dwi"`` inputs,
        and ``limit_to = ["bold"]``, only the "bold" inputs will be parsed. "dwi" will
        be ignored

    participant_label
        Indicate one or more participants to be included from input parsing. This may
        cause errors if subject filters are also specified in pybids_inputs. It may not
        be specified if exclude_participant_label is specified

    exclude_participant_label
        Indicate one or more participants to be excluded from input parsing. This may
        cause errors if subject filters are also specified in pybids_inputs. It may not
        be specified if participant_label is specified

    use_bids_inputs
        If False, returns the classic :class:`BidsDatasetDict` instead of
        :class`BidsDataset`. Setting to True is deprecated as of v0.8, as this is now
        the default behaviour

    index_metadata
        If True indexes metadata of BIDS directory using pybids, otherwise skips
        indexing.

    validate
        If True performs validation of BIDS directory using pybids, otherwise
        skips validation.

    Returns
    -------
    BidsDataset | BidsDatasetDict
        Object containing organized information about the bids inputs for consumption in
        snakemake. See the documentation of :class:`BidsDataset` for details and
        examples.

    Example
    -------
    As an example, consider the following BIDS dataset::

        example
        ├── README.md
        ├── dataset_description.json
        ├── participant.tsv
        ├── sub-001
        │   ├── ses-01
        │   │   ├── anat
        │   │   │   ├── sub-001_ses-01_run-01_T1w.json
        │   │   │   ├── sub-001_ses-01_run-01_T1w.nii.gz
        │   │   │   ├── sub-001_ses-01_run-02_T1w.json
        │   │   │   └── sub-001_ses-01_run-02_T1w.nii.gz
        │   │   └── func
        │   │       ├── sub-001_ses-01_task-nback_bold.json
        │   │       ├── sub-001_ses-01_task-nback_bold.nii.gz
        │   │       ├── sub-001_ses-01_task-rest_bold.json
        │   │       └── sub-001_ses-01_task-rest_bold.nii.gz
        │   └── ses-02
        │       ├── anat
        │       │   ├── sub-001_ses-02_run-01_T1w.json
        │       │   └── sub-001_ses-02_run-01_T1w.nii.gz
        │       └── func
        │           ├── sub-001_ses-02_task-nback_bold.json
        │           ├── sub-001_ses-02_task-nback_bold.nii.gz
        │           ├── sub-001_ses-02_task-rest_bold.json
        │           └── sub-001_ses-02_task-rest_bold.nii.gz
        └── sub-002
            ├── ses-01
            │   ├── anat
            │   │   ├── sub-002_ses-01_run-01_T1w.json
            │   │   ├── sub-002_ses-01_run-01_T1w.nii.gz
            │   │   ├── sub-002_ses-01_run-02_T1w.json
            │   │   └── sub-002_ses-01_run-02_T1w.nii.gz
            │   └── func
            │       ├── sub-002_ses-01_task-nback_bold.json
            │       ├── sub-002_ses-01_task-nback_bold.nii.gz
            │       ├── sub-002_ses-01_task-rest_bold.json
            │       └── sub-002_ses-01_task-rest_bold.nii.gz
            └── ses-02
                └── anat
                    ├── sub-002_ses-02_run-01_T1w.json
                    ├── sub-002_ses-02_run-01_T1w.nii.gz
                    ├── sub-002_ses-02_run-02_T1w.json
                    └── sub-002_ses-02_run-02_T1w.nii.gz

    With the following ``pybids_inputs`` defined in the config file::

        pybids_inputs:
          bold:
            filters:
              suffix: 'bold'
              extension: '.nii.gz'
              datatype: 'func'
            wildcards:
              - subject
              - session
              - acquisition
              - task
              - run

    Then ``generate_inputs(bids_dir, pybids_input)`` would return the following values::

        BidsDataset({
            "bold": BidsComponent(
                name="bold",
                path="bids/sub-{subject}/ses-{session}/func/sub-{subject}_ses-{session}\
_task-{task}_bold.nii.gz",
                zip_lists={
                    "subject": ["001",   "001",  "001",   "001",  "002",   "002" ],
                    "session": ["01",    "01",   "02",    "02",   "01",    "01"  ],
                    "task":    ["nback", "rest", "nback", "rest", "nback", "rest"],
                },
            ),
            "t1w": BidsComponent(
                name="t1w",
                path="example/sub-{subject}/ses-{session}/anat/sub-{subject}_\
ses-{session}_run-{run}_T1w.nii.gz",
                zip_lists={
                    "subject": ["001", "001", "001", "002", "002", "002", "002"],
                    "session": ["01",  "01",  "02",  "01",  "01",  "02",  "02" ],
                    "run":     ["01",  "02",  "01",  "01",  "02",  "01",  "02" ],
                },
            ),
        })
    """
    postfilters = PostFilter()
    postfilters.add_filter("subject", participant_label, exclude_participant_label)

    pybidsdb_dir, pybidsdb_reset = _normalize_database_args(
        pybidsdb_dir, pybidsdb_reset, pybids_database_dir, pybids_reset_database
    )

    # Generates a BIDSLayout
    layout = (
        _gen_bids_layout(
            bids_dir=bids_dir,
            derivatives=derivatives,
            pybids_config=pybids_config,
            pybidsdb_dir=pybidsdb_dir,
            pybidsdb_reset=pybidsdb_reset,
            index_metadata=index_metadata,
            validate=validate,
        )
        if not _all_custom_paths(pybids_inputs)
        else None
    )

    bids_inputs = _get_components(
        bids_layout=layout,
        inputs_config=pybids_inputs,
        limit_to=limit_to,
        postfilters=postfilters,
    )

    if use_bids_inputs is True:
        _logger.warning(
            "The parameter `use_bids_inputs` in generate_inputs() is now set, by "
            "default, to True. Manually setting it to True is deprecated as of version "
            "0.8. "
        )
    elif use_bids_inputs is None:
        use_bids_inputs = True

    try:
        dataset = BidsDataset.from_iterable(bids_inputs, layout)
    except DuplicateComponentError as err:
        msg = (
            f"Multiple components found with the same name: {err.duplicated_names_str}"
        )
        raise ConfigError(msg) from err

    if use_bids_inputs:
        return dataset
    return dataset.as_dict


def _normalize_database_args(
    pybidsdb_dir: Path | str | None,
    pybidsdb_reset: bool | str | None,
    pybids_database_dir: Path | str | None,
    pybids_reset_database: str | bool | None,
) -> tuple[Path | str | None, bool]:
    """Handle deprecated arguments for pybidsdb."""
    if pybids_database_dir is not None:
        warnings.warn(
            "The parameter `pybids_database_dir` in generate_inputs() is deprecated "
            "and will be removed in the next release. To set the pybids database, use "
            "the `pybidsdb_dir` parameter instead.",
            stacklevel=3,
        )
    if pybids_reset_database is not None:
        warnings.warn(
            "The parameter `pybids_reset_database` in generate_inputs() is deprecated "
            "and will be removed in the next release. To reset the pybids database, "
            "use the `pybidsdb_reset` parameter instead.",
            stacklevel=3,
        )

    pybidsdb_dir = pybidsdb_dir or pybids_database_dir
    pybidsdb_reset = (
        pybidsdb_reset
        if pybidsdb_reset is not None
        else pybids_reset_database
        if pybids_reset_database is not None
        else False
    )

    depr_len = len(DEPRECATION_FLAG)
    if (
        pybidsdb_dir is not None
        and str(pybidsdb_dir).startswith(DEPRECATION_FLAG)
        and str(pybidsdb_dir).endswith(DEPRECATION_FLAG)
    ):
        warnings.warn(
            "The config value `pybids_db_dir` is deprecated and will be removed in a "
            "future release. To access a CLI-specified pybids database directory, use "
            '`config.get("pybidsdb_dir")` instead.',
            stacklevel=3,
        )
        pybidsdb_dir = str(pybidsdb_dir)[depr_len:-depr_len]
    try:
        if (
            isinstance(pybidsdb_reset, str)
            and pybidsdb_reset.startswith(DEPRECATION_FLAG)
            and pybidsdb_reset.endswith(DEPRECATION_FLAG)
        ):
            pybidsdb_reset = bool(int(pybidsdb_reset[depr_len:-depr_len]))
            warnings.warn(
                "The config value `pybids_db_reset` is deprecated and will be removed "
                "in a future release. To access CLI-specified pybids database reset "
                'instructions, use `config.get("pybidsdb_reset")` instead.',
                stacklevel=3,
            )
    except ValueError as err:
        msg = "pybidsdb_reset must be a boolean"
        raise TypeError(msg) from err

    if not isinstance(pybidsdb_reset, bool):
        msg = "pybidsdb_reset must be a boolean"
        raise TypeError(msg)

    return pybidsdb_dir, pybidsdb_reset


def _all_custom_paths(config: InputsConfig):
    """Check that all input components have a custom path."""
    return all(comp.get("custom_path") for comp in config.values())


def _is_local_relative(path: Path | str):
    """Test if a path location is local path relative to the current working directory.

    Parameter
    ---------
        path
            A UPath, Path, or str object to be checked

    Returns
    -------
    is_url : bool
        True if the path is relative
    """
    path_str = str(path)
    is_doubleslash_schemed = "://" in path_str
    return not is_doubleslash_schemed and not os.path.isabs(path_str)


def _gen_bids_layout(
    *,
    bids_dir: Path | str,
    derivatives: Path | str | bool,
    pybidsdb_dir: Path | str | None,
    pybidsdb_reset: bool,
    pybids_config: Path | str | None = None,
    index_metadata: bool = False,
    validate: bool = False,
) -> BIDSLayout:
    """Create (or reindex) the BIDSLayout.

    Parameters
    ----------
    bids_dir
        Path to bids directory

    derivatives
        A boolean (or path(s) to derivatives datasets) that
        determines whether snakebids will search in the
        derivatives subdirectory of the input dataset.

    pybidsdb_dir
        Path to database directory. If None is provided, database
        is not used

    pybidsdb_reset
        A boolean that determines whether to reset / overwrite
        existing database.

    index_metadata
        A boolean that determines whether to parse and index metadata

    validate
        A boolean that determines whether to validate the bids dataset

    Returns
    -------
    layout : BIDSLayout
        Layout from pybids for accessing the BIDS dataset to grab paths
    """
    # Check for database_dir
    # If blank, assume db not to be used
    if not pybidsdb_dir:
        pybidsdb_dir = None
    # Otherwise check for relative path and update
    elif not Path(pybidsdb_dir).is_absolute():
        pybidsdb_dir = None
        _logger.warning("Absolute path must be provided, database will not be used")

    return BIDSLayout(
        str(bids_dir),
        derivatives=derivatives,
        validate=validate,
        config=pybids_config,
        database_path=pybidsdb_dir,
        reset_database=pybidsdb_reset,
        indexer=BIDSLayoutIndexer(validate=False, index_metadata=index_metadata),
    )


def write_derivative_json(snakemake: Snakemake, **kwargs: dict[str, Any]) -> None:
    """Update sidecar file with provided sources and parameters.

    Intended for usage in snakemake scripts.

    Parameters
    ----------
    snakemake : struct Snakemake
        structure passed to snakemake python scripts, containing input,
        output, params, config ...
        This function requires input.json and output.json to be defined, as
        it will read and write json files
    """
    sidecar = json.loads(Path(snakemake.input.json).read_text())

    sidecar.update(
        {
            "Sources": [snakemake.input],
            "Parameters": snakemake.params,
            **kwargs,
        }
    )

    with open(snakemake.output.json, "w", encoding="utf-8") as outfile:
        json.dump(sidecar, outfile, indent=4)


def _get_components(
    *,
    bids_layout: BIDSLayout | None,
    inputs_config: InputsConfig,
    postfilters: PostFilter,
    limit_to: Iterable[str] | None = None,
):
    """Generate components based on components config and a bids layout.

    Parameters
    ----------
    bids_layout : BIDSLayout
        Layout from pybids for accessing the BIDS dataset to grab paths.

    inputs_config
        Dictionary indexed by modality name, specifying the filters and
        wildcards for each pybids input.

    limit_to
        List of inputs to skip, this used by snakebids to exclude modalities based on
        cmd-line args.

    postfilters
        Filters to all components after delineation.

    Yields
    ------
    BidsComponent:
        One BidsComponent is yielded for each modality described by ``pybids_inputs``.

    Raises
    ------
    ConfigError
        In response to invalid configuration, missing components, or parsing errors.
    """
    for name in limit_to or inputs_config:
        comp = _get_component(
            bids_layout=bids_layout,
            component=inputs_config[name],
            input_name=name,
            postfilters=postfilters,
        )
        if comp is not None:
            yield comp


def _get_component(
    bids_layout: BIDSLayout | None,
    component: InputConfig,
    *,
    input_name: str,
    postfilters: PostFilter,
) -> BidsComponent | None:
    """Create component based on provided config.

    Parameters
    ----------
    bids_layout
        Layout from pybids for accessing the BIDS dataset to grab paths

    component
        Dictionary indexed by modality name, specifying the filters and
        wildcards for each pybids input.

    input_name
        Name of the component.

    postfilters
        Filters to component after delineation

    Raises
    ------
    ConfigError
        In response to invalid configuration, missing components, or parsing errors.
    """
    _logger.debug("Grabbing inputs for %s...", input_name)

    filters = UnifiedFilter(component, postfilters or {})

    if "custom_path" in component:
        path = component["custom_path"]
        zip_lists = _parse_custom_path(path, filters=filters)
        return BidsComponent(name=input_name, path=path, zip_lists=zip_lists)

    if bids_layout is None:
        msg = (
            f"No valid bids dir given, but {input_name} does not have a "
            "custom_path specified."
        )
        raise RunError(msg)

    zip_lists: dict[str, list[str]] = defaultdict(list)
    paths: set[str] = set()
    try:
        matching_files = get_matching_files(bids_layout, filters)
    except FilterSpecError as err:
        raise err.get_config_error(input_name) from err

    for img in matching_files:
        wildcards: list[str] = [
            wildcard
            for wildcard in set(component.get("wildcards", []))
            if wildcard in img.entities
        ]
        _logger.debug("Wildcards %s found entities for %s", wildcards, img.path)

        try:
            path, parsed_wildcards = _parse_bids_path(img.path, wildcards)
        except BidsParseError as err:
            msg = (
                "Parsing failed:\n"
                f"  Entity: {err.entity.entity}\n"
                f"  Pattern: {err.entity.regex}\n"
                f"  Path: {img.path}\n"
                "\n"
                "Pybids parsed this path using the pattern: "
                f"{bids_layout.entities[err.entity.entity].regex}\n"
                "\n"
                "Snakebids is not currently able to handle this entity. If it is a "
                "custom entity, its `tag-` must be configured to be the same as "
                "its name. Its entry in your pybids config file should look like:\n"
                f'{{\n\t"name": "{err.entity.entity}",\n'
                f'\t"pattern":"{err.entity.entity}-(<value_pattern>)"\n}}\n'
                f"If {err.entity.entity} is an official pybids entity, please "
                "ensure you are using the latest version of snakebids"
            )
            raise ConfigError(msg) from err

        for wildcard_name, value in parsed_wildcards.items():
            zip_lists[wildcard_name].append(value)

        paths.add(path)

    # now, check to see if unique
    if len(paths) == 0:
        _logger.warning(
            "No input files found for snakebids component %s:\n"
            "    filters:\n%s\n"
            "    wildcards:\n%s",
            input_name,
            "\n".join(
                [
                    f"       {key}: {val}"
                    for key, val in component.get("filters", {}).items()
                ]
            ),
            "\n".join(
                [f"       {wildcard}" for wildcard in component.get("wildcards", [])]
            ),
        )
        return None
    try:
        path = itx.one(paths)
    except ValueError as err:
        msg = (
            f"Multiple path templates for one component. Use --filter_{input_name} to "
            f"narrow your search or --wildcards_{input_name} to make the template more "
            "generic.\n"
            f"\tcomponent = {input_name!r}\n"
            f"\tpath_templates = [\n\t\t" + ",\n\t\t".join(map(repr, paths)) + "\n\t]\n"
        ).expandtabs(4)
        raise ConfigError(msg) from err

    if filters.has_empty_postfilter:
        return BidsComponent(
            name=input_name, path=path, zip_lists={key: [] for key in zip_lists}
        )

    return BidsComponent(name=input_name, path=path, zip_lists=zip_lists).filter(
        regex_search=True, **filters.post_exclusions
    )


def _parse_custom_path(
    input_path: Path | str,
    filters: UnifiedFilter,
) -> ZipList:
    """Glob wildcards from a custom path and apply filters.

    This replicates pybids path globbing for any custom path. Input path should have
    wildcards in braces as in "path/of/{wildcard_1}/{wildcard_2}_{wildcard_3}" Output
    will be arranged into a zip list of matches, list of matches, and Snakemake wildcard
    for each wildcard.

    Note that, currently, this will get confused if wildcard content matches
    non-wildcard content. For example, considering the path template above, the example
    "path/of/var1/variable_2_var3" would bug out because of the extra underscore.

    Parameters
    ----------
    input_path : str
        Path to be globbed
    regex_search : bool
        If True, use regex matching for filtering rather than simple equality
    **filters : str or list of str
        Values to keep. Each argument is the name of the entity to search

    Returns
    -------
    input_zip_list, input_list, input_wildcards
    """
    if not (wildcards := glob_wildcards(input_path)):
        _logger.warning("No wildcards defined in %s", input_path)

    # Log an error if no matches found
    if len(itx.first(wildcards.values())) == 0:
        _logger.error("No matching files for %s", input_path)
        return wildcards

    # Return the output values, running filtering on the zip_lists
    result = filter_list(
        wildcards,
        filters.without_bools,
        regex_search=False,
    )
    if not filters.post_exclusions:
        return result

    return filter_list(result, filters.post_exclusions, regex_search=True)


def _parse_bids_path(path: str, entities: Iterable[str]) -> tuple[str, dict[str, str]]:
    """Replace parameters in an bids path with the given wildcard {tags}.

    Parameters
    ----------
    path : str
        BIDS path
        (e.g. "root/sub-01/ses-01/sub-01_ses-01_T1w.nii.gz")
    wildcards : iterable of str
        BIDS entities to replace with wildcards. (e.g. "subject", "session", "suffix")

    Returns
    -------
    path : str
        Original path with the original entities replaced with wildcards.
        (e.g. "root/sub-{subject}/ses-{session}/sub-{subject}_ses-{session}_{suffix}")
    matches : iterable of (wildcard, value)
        The values matched with each wildcard
    """
    # If path is relative, we need to get a slash in front of it to ensure parsing works
    # correctly. So prepend "./" or ".\" and run function again, then strip before
    # returning
    if _is_local_relative(path) and get_first_dir(path) != ".":
        path_, wildcard_values = _parse_bids_path(os.path.join(".", path), entities)
        return str(Path(path_)), wildcard_values

    entities = list(entities)

    matches = sorted(
        (
            (entity, match)
            for entity in map(BidsEntity, entities)
            for match in re.finditer(entity.regex, path)
        ),
        key=lambda match: match[1].start(2),
    )

    wildcard_values: dict[str, str] = {
        entity.wildcard: match.group(2) for entity, match in matches
    }
    if len(wildcard_values) != len(entities):
        unmatched = (
            set(map(BidsEntity, entities))
            .difference(match[0] for match in matches)
            .pop()
        )
        raise BidsParseError(path=path, entity=unmatched)

    num_matches = len(matches)
    new_path: list[str] = []
    for i in range(num_matches + 1):
        start = matches[i - 1][1].end(2) if i else 0
        end = len(path) if i == num_matches else matches[i][1].start(2)
        # Pybids technically allows `{` in the extension, so we escape it
        new_path.append(path[start:end].replace("{", "{{").replace("}", "}}"))
        if i < num_matches:
            new_path.append(f"{{{matches[i][0].wildcard}}}")
    return "".join(new_path), wildcard_values


def get_wildcard_constraints(image_types: InputsConfig) -> dict[str, str]:
    """Return a wildcard_constraints dict for use in snakemake.

    Contains constraints for all wildcards in the dynamically grabbed inputs.

    Parameters
    ----------
    image_types
        Component configuration dict

    Returns
    -------
        Dict containing wildcard constraints for all wildcards in the
        inputs, with typical bids naming constraints, ie letters and numbers
        ``[a-zA-Z0-9]+``.
    """
    bids_constraints = "[a-zA-Z0-9]+"
    return {
        entity: bids_constraints
        for imgtype in image_types
        for entity in image_types[imgtype]
    }



from __future__ import annotations

import functools as ft
import itertools as it
import textwrap
import warnings
from collections.abc import Iterable
from math import inf
from pathlib import Path
from string import Formatter
from typing import Any, NoReturn, cast, overload

import attr
import more_itertools as itx
from bids import BIDSLayout
from pvandyken.deprecated import deprecated
from typing_extensions import Self, TypedDict

import snakebids.utils.sb_itertools as sb_it
from snakebids.core.filtering import filter_list
from snakebids.exceptions import DuplicateComponentError
from snakebids.io.console import get_console_size
from snakebids.io.printing import format_zip_lists, quote_wrap
from snakebids.snakemake_compat import expand as sn_expand
from snakebids.types import ZipList
from snakebids.utils.containers import ImmutableList, MultiSelectDict
from snakebids.utils.utils import get_wildcard_dict, property_alias, zip_list_eq


class BidsDatasetDict(TypedDict):
    """Dict equivalent of BidsInputs, for backwards-compatibility."""

    input_path: dict[str, str]
    input_zip_lists: dict[str, dict[str, list[str]]]
    input_lists: dict[str, dict[str, list[str]]]
    input_wildcards: dict[str, dict[str, str]]
    subjects: list[str]
    sessions: list[str]
    subj_wildcards: dict[str, str]


class BidsComponentRow(ImmutableList[str]):
    """A single row from a BidsComponent.

    This class is derived by indexing a single entity from a :class:`BidsComponent` or
    :class:`BidsPartialComponent`. It should not be constructed manually.

    The class is a subclass of :class:`ImmutableList` and can thus be treated as a
    tuple. Indexing it via ``row[<int>]`` gives the entity-value of the selected entry.

    The :attr:`~BidsComponentRow.entities` and :attr:`~BidsComponentRow.wildcards`
    directly return the list of unique entity-values or the ``{brace-wrapped-entity}``
    name corresponding to the row, rather than a dict.

    The :meth:`~BidsComponentRow.expand` and :meth:`~BidsComponentRow.filter` methods
    behave as they would in a :class:`BidsComponent` with a single entity.

    """

    def __init__(self, iterable: Iterable[str], /, entity: str):
        super().__init__(iterable)
        self.entity = entity

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({list(self._data)}, entity="{self.entity}")'

    @property
    def entities(self) -> tuple[str, ...]:
        """The unique values associated with the component."""
        return tuple(set(self._data))

    @property
    def wildcards(self) -> str:
        """The entity name wrapped in wildcard braces."""
        return f"{{{self.entity}}}"

    @property
    def zip_lists(self) -> ZipList:
        """Table of entities and values.

        Each key is a wildcard entity and each value is a list of the
        values found for that entity. Each of these lists has length equal to the number
        of images matched for this modality, so they can be zipped together to get a
        list of the wildcard values for each file.
        """
        return MultiSelectDict({self.entity: list(self._data)})

    def __eq__(self, other: BidsComponentRow | object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return super().__eq__(other)

    def expand(
        self,
        paths: Iterable[Path | str] | Path | str,
        /,
        allow_missing: bool | str | Iterable[str] = False,
        **wildcards: str | Iterable[str],
    ) -> list[str]:
        """Safely expand over given paths with component wildcards.

        Uses the entity-values represented by this row to expand over the given paths.
        Extra wildcards can be specified as keyword arguments.

        By default, expansion over paths with extra wildcards not accounted for by the
        component causes an error. This prevents accidental partial expansion. To allow
        the passage of extra wildcards without expansion,set ``allow_missing`` to
        ``True``.

        Uses the snakemake :ref:`expand <snakemake:snakefiles_expand>` under the hood.

        Parameters
        ----------
        paths:
            Path or list of paths to expand over
        allow_missing:
            If True, allow ``{wildcards}`` in the provided paths that are not present
            either in the component or in the extra provided ``**wildcards``. These
            wildcards will be preserved in the returned paths.
        wildcards:
            Each keyword should be the name of an wildcard in the provided paths.
            Keywords not found in the path will be ignored. Keywords take values or
            lists of values to be expanded over the provided paths.
        """
        return sn_expand(
            list(itx.always_iterable(paths)),
            allow_missing=allow_missing
            if isinstance(allow_missing, bool)
            else list(itx.always_iterable(allow_missing)),
            **{self.entity: list(dict.fromkeys(self._data))},
            **{
                wildcard: list(itx.always_iterable(v))
                for wildcard, v in wildcards.items()
            },
        )

    def filter(
        self,
        spec: str | Iterable[str] | None = None,
        /,
        *,
        regex_search: bool | str | Iterable[str] = False,
        **filters: str | Iterable[str],
    ) -> Self:
        """Filter component based on provided entity filters.

        Extracts a subset of the entity-values present in the row.

        Takes entities as keyword arguments assigned to values or list of values to
        select from the component. Only columns containing the provided entity-values
        are kept. If no matches are found, a component with the all the original
        entities but with no values will be returned.

        Returns a brand new :class:`~snakebids.BidsComponentRow`. The original component
        is not modified.

        Parameters
        ----------
        spec
            Value or iterable of values assocatiated with the ComponentRow's
            :attr:`~snakebids.BidsComponentRow.entity`. Equivalent to specifying
            ``.filter(entity=value)``
        regex_search
            Treat filters as regex patterns when matching with entity-values.
        filters
            Keyword-value(s) filters as in :meth:`~snakebids.BidsComponent.filter`.
            Here, the only valid filter is the
            :attr:`~snakebids.BidsComponentRow.entity` of the
            :class:`~snakebids.BidsComponentRow`; all others will be ignored.
        """
        if not isinstance(regex_search, bool):
            msg = "regex_search must be a boolean"
            raise TypeError(msg)
        if spec is not None:
            if filters:
                msg = "Both __spec and filters cannot be used simultaneously"
                raise ValueError(msg)
            filters = {self.entity: spec}
        entity, data = itx.first(
            filter_list(
                {self.entity: self._data}, filters, regex_search=regex_search
            ).items()
        )
        return self.__class__(data, entity=entity)


@attr.define(kw_only=True)
class BidsPartialComponent:
    """Primitive representation of a bids data component.

    See :class:`BidsComponent` for an extended definition of a data component.

    ``BidsPartialComponents`` are typically derived from a :class:`BidsComponent`. They
    do not store path information, and do not represent real data files. They just have
    a table of entity-values, typically a subset of those present in their source
    :class:`BidsComponent`.

    Despite this, ``BidsPartialComponents`` still allow you to expand the data table
    over new paths, allowing you to derive paths from your source dataset.

    The members of ``BidsPartialComponent`` are identical to :class:`BidsComponent` with
    the following exceptions:

    - No :attr:`~BidsComponent.name` or :attr:`~BidsComponent.path`
    - :meth:`~BidsPartialComponent.expand` must be given a path or list of paths as the
      first argument

    ``BidsPartialComponents`` are immutable: their values cannot be altered.
    """

    _zip_lists: ZipList = attr.field(
        on_setattr=attr.setters.frozen, converter=MultiSelectDict, alias="zip_lists"
    )

    @_zip_lists.validator  # type: ignore
    def _validate_zip_lists(self, __attr: str, value: dict[str, list[str]]) -> None:
        lengths = {len(val) for val in value.values()}
        if len(lengths) > 1:
            msg = "zip_lists must all be of equal length"
            raise ValueError(msg)

    def __repr__(self) -> str:
        return self.pformat()

    @overload
    def __getitem__(self, key: str, /) -> BidsComponentRow: ...

    @overload
    def __getitem__(self, key: tuple[str, ...], /) -> BidsPartialComponent: ...

    def __getitem__(
        self, key: str | tuple[str, ...], /
    ) -> BidsComponentRow | BidsPartialComponent:
        if isinstance(key, tuple):
            # Use dict.fromkeys for de-duplication
            return BidsPartialComponent(
                zip_lists={key: self.zip_lists[key] for key in dict.fromkeys(key)}
            )
        return BidsComponentRow(self.zip_lists[key], entity=key)

    def __bool__(self) -> bool:
        """Truth of a BidsComponent is based on whether it has values.

        It is not based on whether it has any entities. This is because
        :meth:`~BidsPartialComponent.filter` returns a component retaining all entities
        but with no values, making that the standard for emptiness. It also makes it
        consistent with :class:`BidsComponentRow`, which always has an entity name
        stored, but may or may not have values.
        """
        return bool(itx.first(self.zip_lists))

    def _pformat_body(self) -> None | str | list[str]:
        """Extra properties to be printed within pformat.

        Meant for implementation in subclasses
        """
        return None

    def pformat(self, max_width: int | float | None = None, tabstop: int = 4) -> str:
        """Pretty-format component.

        Parameters
        ----------
        max_width
            Maximum width of characters for output. If possible, zip_list table will be
            elided to fit within this width
        tabstop
            Number of spaces for output indentation
        """
        width = max_width or get_console_size()[0] or inf
        body = it.chain(
            itx.always_iterable(self._pformat_body() or []),
            [
                "zip_lists="
                f"{format_zip_lists(self.zip_lists, width - tabstop, tabstop)},",
            ],
        )
        output = [
            f"{self.__class__.__name__}(",
            textwrap.indent("\n".join(body), " " * tabstop),
            ")",
        ]
        return "\n".join(output)

    # Note: we can't use cached property here because it's incompatible with slots.
    _input_lists: MultiSelectDict[str, list[str]] | None = attr.field(
        default=None, init=False, eq=False, repr=False
    )
    _input_wildcards: MultiSelectDict[str, str] | None = attr.field(
        default=None, init=False, eq=False, repr=False
    )
    _entities: list[str] | None = attr.field(
        default=None, init=False, eq=False, repr=False
    )

    @property
    def zip_lists(self):
        """Table of unique wildcard groupings for each member in the component.

        Dictionary where each key is a wildcard entity and each value is a list of the
        values found for that entity. Each of these lists has length equal to the number
        of images matched for this modality, so they can be zipped together to get a
        list of the wildcard values for each file.
        """
        return self._zip_lists

    @property
    def entities(self) -> MultiSelectDict[str, list[str]]:
        """Component entities and their associated values.

        Dictionary where each key is an entity and each value is a list of the
        unique values found for that entity. These lists might not be the same length.
        """
        if self._input_lists is None:
            self._input_lists = MultiSelectDict(
                {entity: list(set(values)) for entity, values in self.zip_lists.items()}
            )
        return self._input_lists

    @property
    def wildcards(self) -> MultiSelectDict[str, str]:
        """Wildcards in brace-wrapped syntax.

        Dictionary where each key is the name of a wildcard entity, and each value is
        the Snakemake wildcard used for that entity.
        """
        if self._input_wildcards is None:
            self._input_wildcards = MultiSelectDict(get_wildcard_dict(self.zip_lists))
        return self._input_wildcards

    @property
    def input_zip_lists(self) -> ZipList:
        """Alias of :attr:`zip_lists <snakebids.BidsComponent.zip_lists>`.

        Dictionary where each key is a wildcard entity and each value is a list of the
        values found for that entity. Each of these lists has length equal to the number
        of images matched for this modality, so they can be zipped together to get a
        list of the wildcard values for each file.
        """
        return self.zip_lists

    @property_alias(entities, "entities", "snakebids.BidsComponent.entities")
    def input_lists(self):
        return self.entities

    @property_alias(wildcards, "wildcards", "snakebids.BidsComponent.wildcards")
    def input_wildcards(self):
        return self.wildcards

    def __eq__(self, other: BidsComponent | object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return zip_list_eq(self.zip_lists, other.zip_lists)

    def expand(
        self,
        paths: Iterable[Path | str] | Path | str,
        /,
        allow_missing: bool | str | Iterable[str] = False,
        **wildcards: str | Iterable[str],
    ) -> list[str]:
        """Safely expand over given paths with component wildcards.

        Uses the entity-value combinations found in the dataset to expand over the given
        paths. Extra wildcards can be specified as keyword arguments.

        By default, expansion over paths with extra wildcards not accounted for by the
        component causes an error. This prevents accidental partial expansion. To allow
        the passage of extra wildcards without expansion,set ``allow_missing`` to
        ``True``.

        Uses the snakemake :ref:`expand <snakemake:snakefiles_expand>` under the hood.

        Parameters
        ----------
        paths:
            Path or list of paths to expand over
        allow_missing:
            If True, allow ``{wildcards}`` in the provided paths that are not present
            either in the component or in the extra provided ``**wildcards``. These
            wildcards will be preserved in the returned paths.
        wildcards:
            Each keyword should be the name of an wildcard in the provided paths.
            Keywords not found in the path will be ignored. Keywords take values or
            lists of values to be expanded over the provided paths.
        """

        def sequencify(item: bool | str | Iterable[str]) -> bool | list[str]:
            if isinstance(item, bool):
                return item
            return list(itx.always_iterable(item))

        allow_missing_seq = sequencify(allow_missing)
        if self.zip_lists:
            inner_expand = list(
                # order preserving deduplication
                dict.fromkeys(
                    sn_expand(
                        list(itx.always_iterable(paths)),
                        zip,
                        allow_missing=True if wildcards else allow_missing_seq,
                        **self.zip_lists,
                    )
                )
            )
        else:
            inner_expand = list(itx.always_iterable(paths))
        if not wildcards:
            return inner_expand

        return sn_expand(
            inner_expand,
            allow_missing=allow_missing_seq,
            # Turn all the wildcard items into lists because Snakemake doesn't handle
            # iterables very well
            **{
                wildcard: list(itx.always_iterable(v))
                for wildcard, v in wildcards.items()
            },
        )

    def filter(
        self,
        *,
        regex_search: bool | str | Iterable[str] = False,
        **filters: str | Iterable[str],
    ) -> Self:
        """Filter component based on provided entity filters.

        This method allows you to expand over a subset of your wildcards. This could be
        useful for extracting subjects from a specific patient group, running different
        rules on different acquisitions, and any other reason you may need to filter
        your data after the workflow has already started.

        Takes entities as keyword arguments assigned to values or list of values to
        select from the component. Only columns containing the provided entity-values
        are kept. If no matches are found, a component with the all the original
        entities but with no values will be returned.

        Returns a brand new :class:`~snakebids.BidsComponent`. The original component is
        not modified.

        Parameters
        ----------
        regex_search
            Treat filters as regex patterns when matching with entity-values.
        filters
            Each keyword should be the name of an entity in the component. Entities not
            found in the component will be ignored. Keywords take values or a list of
            values to be matched with the component
            :attr:`~snakebids.BidsComponent.zip_lists`
        """
        if not isinstance(regex_search, bool):
            msg = "regex_search must be a boolean"
            raise TypeError(msg)
        if not filters:
            return self
        return attr.evolve(
            self,
            zip_lists=filter_list(self.zip_lists, filters, regex_search=regex_search),
        )


@attr.define(kw_only=True)
class BidsComponent(BidsPartialComponent):
    """Representation of a bids data component.

    A component is a set of data entries all corresponding to the same type of object.
    Entries vary over a set of entities. For example, a component may represent all the
    unprocessed, T1-weighted anatomical images acquired from a group of 100 subjects,
    across 2 sessions, with three runs per session. Here, the subject, session, and run
    are the entities over which the component varies. Each entry in the component has
    a single value assigned for each of the three entities (e.g subject 002, session
    01, run 1).

    Each entry can be defined solely by its wildcard values. The complete collection of
    entries can thus be stored as a table, where each row represents an entity and each
    column represents an entry.

    ``BidsComponent`` stores and indexes this table. It uses 'row-first' indexing,
    meaning first an entity is selected, then an entry. It also has a number of
    properties and methods making it easier to incorporate the data in a snakemake
    workflow.

    In addition, ``BidsComponent`` stores a template :attr:`~BidsComponent.path <path>`
    derived from the source dataset. This path is used by the
    :meth:`~BidsComponent.expand` method to recreate the original filesystem paths.

    The real power of the ``BidsComponent``, however, is in creating derived paths based
    on the original dataset. Using the :meth`~BidsComponent.expand` method, you can pass
    new paths with ``{wildcard}`` placeholders wrapped in braces and named according to
    the entities in the component. These placeholders will be substituted with the
    entity values saved in the table, giving you a list of paths the same length as the
    number of entries in the component.

    BidsComponents are immutable: their values cannot be altered.
    """

    name: str = attr.field(on_setattr=attr.setters.frozen)
    """Name of the component"""

    path: str = attr.field(on_setattr=attr.setters.frozen)
    """Wildcard-filled path that matches the files for this component."""

    _zip_lists: ZipList = attr.field(
        on_setattr=attr.setters.frozen, converter=MultiSelectDict, alias="zip_lists"
    )

    @_zip_lists.validator  # type: ignore
    def _validate_zip_lists(self, __attr: str, value: dict[str, list[str]]) -> None:
        super()._validate_zip_lists(__attr, value)
        _, raw_fields, *_ = sb_it.unpack(
            zip(*Formatter().parse(self.path), strict=True), [[], [], []]
        )
        raw_fields = cast("Iterable[str]", raw_fields)
        if (fields := set(filter(None, raw_fields))) != set(value):
            msg = (
                "zip_lists entries must match the wildcards in input_path: "
                f"{self.path}: {fields} != zip_lists: {set(value)}"
            )
            raise ValueError(msg)

    def __repr__(self) -> str:
        return self.pformat()

    def _pformat_body(self):
        return [
            f"name={quote_wrap(self.name)},",
            f"path={quote_wrap(self.path)},",
        ]

    def __eq__(self, other: BidsComponent | object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        if self.name != other.name:
            return False

        if self.path != other.path:
            return False

        return super().__eq__(other)

    def expand(
        self,
        paths: Iterable[Path | str] | Path | str | None = None,
        /,
        allow_missing: bool | str | Iterable[str] = False,
        **wildcards: str | Iterable[str],
    ) -> list[str]:
        """Safely expand over given paths with component wildcards.

        Uses the entity-value combinations found in the dataset to expand over the given
        paths. If no path is provided, expands over the component
        :attr:`~snakebids.BidsComponent.path` (thus returning the original files used to
        create the component). Extra wildcards can be specified as keyword arguments.

        By default, expansion over paths with extra wildcards not accounted for by the
        component causes an error. This prevents accidental partial expansion. To allow
        the passage of extra wildcards without expansion,set ``allow_missing`` to
        ``True``.

        Uses the snakemake :ref:`expand <snakemake:snakefiles_expand>` under the hood.

        Parameters
        ----------
        paths:
            Path or list of paths to expand over. If not provided, the component's own
            :attr:`~BidsComponent.path` will be expanded over.
        allow_missing:
            If True, allow ``{wildcards}`` in the provided paths that are not present
            either in the component or in the extra provided ``**wildcards``. These
            wildcards will be preserved in the returned paths.
        wildcards:
            Each keyword should be the name of an wildcard in the provided paths.
            Keywords not found in the path will be ignored. Keywords take values or
            lists of values to be expanded over the provided paths.
        """
        paths = paths or self.path
        return super().expand(paths, allow_missing, **wildcards)

    @property
    def input_name(self) -> str:
        """Alias of :attr:`name <snakebids.BidsComponent.name>`.

        Name of the component
        """
        return self.name

    @property
    def input_path(self) -> str:
        """Alias of :attr:`path <snakebids.BidsComponent.path>`.

        Wildcard-filled path that matches the files for this component.
        """
        return self.path


class BidsDataset(dict[str, BidsComponent]):
    """A bids dataset parsed by pybids, organized into BidsComponents.

    BidsDatasets are typically generated using :func:`generate_inputs()
    <snakebids.generate_inputs>`, which reads the ``pybids_inputs`` field in your
    snakemake config file and, for each entry, creates a BidsComponent using the
    provided name, wildcards, and filters.

    Individual components can be accessed using bracket-syntax: (e.g.
    ``inputs["t1w"]``).

    Provides access to summarizing information, for instance, the set of all subjects or
    sessions found in the dataset.
    """

    layout: BIDSLayout | None
    """
    Underlying layout generated from pybids. Note that this will be set to None if
    custom paths are used to generate every :class:`component <BidsComponent>`
    """

    def __init__(self, data: Any, layout: BIDSLayout | None = None) -> None:
        super().__init__(data)
        self.layout = layout

    def __getitem__(self, key: str) -> BidsComponent:
        try:
            return super().__getitem__(key)
        except KeyError as err:
            if key in {
                "input_path",
                "input_zip_lists",
                "input_lists",
                "input_wildcards",
                "subjects",
                "sessions",
                "subj_wildcards",
            }:
                msg = (
                    "As of v0.8, generate_inputs() no longer returns a dict by "
                    f"default, but an instance of BidsDataset. As such, '{key}' can no "
                    "longer be accessed via brackets '[]' as before. The original dict "
                    "can be returned by setting `use_bids_inputs` to False in the call "
                    "to generate_inputs(). However, we encourage you to transition to "
                    "the use of `BidsDataset` for long term support"
                )
                raise KeyError(msg) from err
            raise

    def __setitem__(self, _: Any, __: Any) -> NoReturn:
        msg = f"Modification of {self.__class__.__name__} is not yet supported"
        raise NotImplementedError(msg)

    def __repr__(self) -> str:
        return self.pformat()

    def pformat(self, max_width: int | float | None = None, tabstop: int = 4) -> str:
        """Pretty-format dataset.

        Parameters
        ----------
        max_width
            Maximum width of characters for output. If possible, zip_list table will be
            elided to fit within this width
        tabstop
            Number of spaces for output indentation
        """
        width = max_width or get_console_size()[0] or inf
        body = [
            f"{quote_wrap(name)}: {comp.pformat(width - tabstop, tabstop)},"
            for name, comp in self.items()
        ]
        output = [
            "BidsDataset({",
            textwrap.indent("\n".join(body), " " * tabstop),
            "})",
        ]
        return "\n".join(output)

    @ft.cached_property
    @deprecated(
        details="""
        The behaviour of path will change in an upcoming release, where it will refer
        instead to the root path of the dataset. Please access component paths using
        :attr:`Dataset[\\<component_name\\>].path <BidsComponent.path>`
        """,
        deprecated_in="0.8.0",
        admonition="warning",
    )
    def path(self) -> dict[str, str]:
        """Dict mapping :class:`BidsComponent` names to their ``paths``."""
        return {key: value.path for key, value in self.items()}

    @ft.cached_property
    @deprecated(
        details="""
        The behaviour of zip_lists will change in an upcoming release, where it will
        refer instead to the consensus of entity groups across all components in the
        dataset. Please access component zip_lists using
        :attr:`Dataset[\\<component_name\\>].zip_lists <BidsComponent.zip_lists>`
        """,
        deprecated_in="0.8.0",
        admonition="warning",
    )
    def zip_lists(self) -> dict[str, ZipList]:
        """Dict mapping :class:`BidsComponent` names to their ``zip_lists``."""
        return {key: value.zip_lists for key, value in self.items()}

    @ft.cached_property
    @deprecated(
        details="""
        The behaviour of entities will change in the 1.0 release, where it will refer
        instead to the union of all entity-values across all components in the dataset.
        Please access component entity lists using
        :attr:`Dataset[\\<component_name\\>].entities <BidsComponent.entities>`
        """,
        deprecated_in="0.8.0",
        admonition="warning",
    )
    def entities(self) -> dict[str, MultiSelectDict[str, list[str]]]:
        """Dict mapping :class:`BidsComponent` names to their :attr:`~BidsComponent.entities`."""  # noqa: E501
        return {key: value.entities for key, value in self.items()}

    @ft.cached_property
    @deprecated(
        details="""
        The behaviour of wildcards will change in an upcoming release, where it will
        refer instead to the union of all entity-wildcard mappings across all components
        in the dataset. Please access component wildcards using
        :attr:`Dataset[\\<component_name\\>].wildcards <BidsComponent.wildcards>`
        """,
        deprecated_in="0.8.0",
        admonition="warning",
    )
    def wildcards(self) -> dict[str, MultiSelectDict[str, str]]:
        """Dict mapping :class:`BidsComponent` names to their :attr:`~BidsComponent.wildcards`."""  # noqa: E501
        return {key: value.input_wildcards for key, value in self.items()}

    @ft.cached_property
    def subjects(self) -> list[str]:
        """A list of the subjects in the dataset."""
        return [
            *{
                *it.chain.from_iterable(
                    component.entities.get("subject", []) for component in self.values()
                )
            }
        ]

    @ft.cached_property
    def sessions(self) -> list[str]:
        """A list of the sessions in the dataset."""
        return [
            *{
                *it.chain.from_iterable(
                    component.entities.get("session", []) for component in self.values()
                )
            }
        ]

    @ft.cached_property
    def subj_wildcards(self) -> dict[str, str]:
        """The subject and session wildcards applicable to this dataset.

        ``{"subject":"{subject}"}`` if there is only one session, ``{"subject":
        "{subject}", "session": "{session}"}`` if there are multiple sessions.
        """
        if len(self.sessions) == 0:
            return {"subject": "{subject}"}
        return {
            "subject": "{subject}",
            "session": "{session}",
        }

    @property_alias(path, "path", "snakebids.BidsDataset.path")
    def input_path(self) -> dict[str, str]:
        return self.path

    @property_alias(entities, "entities", "snakebids.BidsDataset.entities")
    def input_lists(self) -> dict[str, MultiSelectDict[str, list[str]]]:
        return self.entities

    @property_alias(zip_lists, "zip_lists", "snakebids.BidsDataset.zip_lists")
    def input_zip_lists(self) -> dict[str, MultiSelectDict[str, list[str]]]:
        return self.zip_lists

    @property_alias(wildcards, "wildcards", "snakebids.BidsDataset.wildcards")
    def input_wildcards(self) -> dict[str, MultiSelectDict[str, str]]:
        return self.wildcards

    @property
    def as_dict(self) -> BidsDatasetDict:
        """Get the layout as a legacy dict.

        Included primarily for backward compatibility with older versions of snakebids,
        where generate_inputs() returned a dict rather than the `BidsDataset` class

        Returns
        -------
        BidsDatasetDict
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            return BidsDatasetDict(
                input_path=self.input_path,
                input_lists={k: dict(val) for k, val in self.entities.items()},
                input_wildcards={
                    k: dict(val) for k, val in self.input_wildcards.items()
                },
                input_zip_lists={
                    k: dict(val) for k, val in self.input_zip_lists.items()
                },
                subjects=self.subjects,
                sessions=self.sessions,
                subj_wildcards=self.subj_wildcards,
            )

    @classmethod
    def from_iterable(
        cls, iterable: Iterable[BidsComponent], layout: BIDSLayout | None = None
    ) -> BidsDataset:
        """Construct Dataset from iterable of BidsComponents.

        Parameters
        ----------
        iterable : Iterable[BidsComponent]

        Returns
        -------
        BidsDataset
        """
        components = list(iterable)
        indexed = {bidsinput.name: bidsinput for bidsinput in components}
        if not len(components) == len(indexed):
            raise DuplicateComponentError(
                list(itx.duplicates_everseen([c.name for c in components]))
            )
        return cls(indexed, layout)



import jinja2
from jinja2.ext import Extension

from snakebids.jinja2_ext.toml_encode import toml_string


def format_poetry(item: str):
    """Format a pip style dependency specification for a poetry pyproject.toml.

    Only supports urls (prefixed with @) and version specifiers. No markers or extras.
    The package name should already be stripped.
    """
    if item.strip().startswith("@"):
        url = item[1:].strip()
        if url.startswith("git+"):
            return f"{{ git = {toml_string(url[4:])} }}"
        if url.startswith("file://"):
            return f"{{ path = {toml_string(url[7:])} }}"
        return f"{{ url = {toml_string(url)} }}"
    return toml_string(item)


class FormatDepSpec(Extension):
    """Enable the toml_string filter, which encodes provided value as toml."""

    def __init__(self, env: jinja2.Environment):
        env.filters["format_poetry"] = format_poetry  # type: ignore



from __future__ import annotations

import itertools as it
from argparse import ArgumentParser
from collections.abc import Iterable
from pathlib import Path

import pytest
from hypothesis import given
from hypothesis import strategies as st

from snakebids.exceptions import ConfigError
from snakebids.plugins.bidsargs import BidsArgs
from tests import strategies as sb_st


class _St:
    positional_args = st.text().filter(lambda s: not s.startswith("-"))
    analysis_level_choices = positional_args | st.iterables(positional_args)


class TestAddCliArguments:
    @given(
        analysis_level_choices=_St.positional_args,
        analysis_level_choices_config=st.text(),
    )
    def test_analysis_level_attributes_are_mutually_exclusive(
        self,
        analysis_level_choices: str | Iterable[str],
        analysis_level_choices_config: str,
    ):
        with pytest.raises(ConfigError, match="cannot be simultaneously defined"):
            BidsArgs(
                analysis_level=True,
                analysis_level_choices=analysis_level_choices,
                analysis_level_choices_config=analysis_level_choices_config,
            ).add_cli_arguments(ArgumentParser(), {}, {})

    @given(analysis_level_choices=_St.positional_args)
    def test_analysis_levels_can_be_directly_defined(
        self, analysis_level_choices: str | Iterable[str]
    ):
        parser = ArgumentParser()
        a1, a2 = it.tee(analysis_level_choices)
        bidsargs = BidsArgs(analysis_level=True, analysis_level_choices=a1)
        bidsargs.add_cli_arguments(parser, {}, {})
        for arg in a2:
            nspc = parser.parse_args(["...", "...", arg])
            assert nspc.analysis_level == arg

    @given(
        analysis_level_choices=_St.positional_args,
        analysis_level_choices_config=st.text(),
    )
    def test_analysis_levels_can_be_defined_in_config(
        self,
        analysis_level_choices: str | Iterable[str],
        analysis_level_choices_config: str,
    ):
        parser = ArgumentParser()
        choices = list(analysis_level_choices)
        bidsargs = BidsArgs(
            analysis_level=True,
            analysis_level_choices_config=analysis_level_choices_config,
        )
        bidsargs.add_cli_arguments(parser, {}, {analysis_level_choices_config: choices})
        for arg in choices:
            nspc = parser.parse_args(["...", "...", arg])
            assert nspc.analysis_level == arg

    @given(
        derivatives=st.lists(
            st.text(sb_st.path_characters).filter(lambda s: not s.startswith("-")),
            min_size=1,
        )
    )
    def test_derivatives_converted_to_paths(self, derivatives: list[str]):
        parser = ArgumentParser()
        bidsargs = BidsArgs()
        bidsargs.add_cli_arguments(parser, {}, {})
        nspc = parser.parse_args(
            ["...", "...", "participant", "--derivatives", *derivatives]
        )
        assert nspc.derivatives == [Path(p) for p in derivatives]

    def test_derivatives_true_if_no_paths_given(self):
        parser = ArgumentParser()
        bidsargs = BidsArgs()
        bidsargs.add_cli_arguments(parser, {}, {})
        nspc = parser.parse_args(["...", "...", "participant", "--derivatives"])
        assert nspc.derivatives is True

    def test_derivatives_false_by_default(self):
        parser = ArgumentParser()
        bidsargs = BidsArgs()
        bidsargs.add_cli_arguments(parser, {}, {})
        nspc = parser.parse_args(["...", "...", "participant"])
        assert nspc.derivatives is False



"""
This type stub file was generated by pyright.
"""

from ruamel.yaml.reader import Reader
from ruamel.yaml.scanner import RoundTripScanner, Scanner
from ruamel.yaml.parser import Parser, RoundTripParser
from ruamel.yaml.composer import Composer
from ruamel.yaml.constructor import (
    BaseConstructor,
    Constructor,
    RoundTripConstructor,
    SafeConstructor,
)
from ruamel.yaml.resolver import VersionedResolver

if False: ...
__all__ = ["BaseLoader", "SafeLoader", "Loader", "RoundTripLoader"]

class BaseLoader(Reader, Scanner, Parser, Composer, BaseConstructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class SafeLoader(Reader, Scanner, Parser, Composer, SafeConstructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class Loader(Reader, Scanner, Parser, Composer, Constructor, VersionedResolver):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...

class RoundTripLoader(
    Reader,
    RoundTripScanner,
    RoundTripParser,
    Composer,
    RoundTripConstructor,
    VersionedResolver,
):
    def __init__(
        self,
        stream: StreamTextType,
        version: Optional[VersionType] = ...,
        preserve_quotes: Optional[bool] = ...,
    ) -> None: ...



"""
This type stub file was generated by pyright.
"""

""" Utilities for manipulating package-level settings. """
__all__ = ["set_option", "set_options", "get_option"]
_config_name = ...
conf_path = ...
_default_settings = ...

def set_option(key, value):  # -> None:
    """Set a package-wide option.

    Args:
        key (str): The name of the option to set.
        value (object): The new value of the option.
    """
    ...

def set_options(**kwargs):  # -> None:
    """Set multiple package-wide options.

    Args:
        kwargs: Keyword arguments to pass onto set_option().
    """
    ...

def get_option(key):  # -> dict[str, str] | bool:
    """Retrieve the current value of a package-wide option.

    Args:
        key (str): The name of the option to retrieve.

    """
    ...

def from_file(filenames, error_on_missing=...):  # -> None:
    """Load package-wide settings from specified file(s).

    Args:
        filenames (str, list): Filename or list of filenames containing JSON
            dictionary of settings.
        error_on_missing (bool): If True, raises an error if a file doesn't
            exist.
    """
    ...

def reset_options(update_from_file=...):  # -> None:
    """Reset all options to the package defaults.

    Args:
        update_from_file (bool): If True, re-applies any config files found in
            standard locations.
    """
    ...

_settings = ...



from __future__ import annotations

import argparse
import re
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path

import more_itertools as itx
import pytest
from hypothesis import given
from hypothesis import strategies as st
from pathvalidate import Platform, is_valid_filename
from pytest_mock.plugin import MockerFixture

import snakebids
from snakebids.admin import copier, create_descriptor, gen_parser
from tests.helpers import allow_function_scoped


@pytest.fixture
def parser():
    return gen_parser()


def test_fails_if_no_subcommand(parser: ArgumentParser, mocker: MockerFixture):
    mocker.patch.object(sys, "argv", ["snakebids"])
    with pytest.raises(SystemExit):
        parser.parse_args()


def test_fails_if_invalid_subcommand(parser: ArgumentParser, mocker: MockerFixture):
    mocker.patch.object(sys, "argv", ["snakebids", "dummy"])
    with pytest.raises(SystemExit):
        parser.parse_args()


class TestCreateCommand:
    @given(
        name=st.text()
        .filter(lambda s: not re.match(r"^[a-zA-Z_][a-zA-Z_0-9]*$", s))
        .filter(lambda s: is_valid_filename(s, Platform.LINUX))
        .filter(lambda s: s not in {".", ".."})
    )
    @allow_function_scoped
    def test_create_fails_with_invalid_filename(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        name: str,
        tmp_path: Path,
        capsys: pytest.CaptureFixture[str],
    ):
        mocker.patch.object(sys, "argv", ["snakebids", "create", str(tmp_path / name)])
        args = parser.parse_args()
        with pytest.raises(SystemExit):
            args.func(args)
        capture = capsys.readouterr()
        assert "valid python module" in capture.err
        assert name in capture.err

    @given(
        name=st.text()
        .filter(lambda s: is_valid_filename(s, Platform.LINUX))
        .filter(lambda s: s not in {".", ".."})
    )
    @allow_function_scoped
    def test_create_fails_missing_parent_dir(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        name: str,
        tmp_path: Path,
        capsys: pytest.CaptureFixture[str],
    ):
        path = tmp_path / name / "sub"
        mocker.patch.object(sys, "argv", ["snakebids", "create", str(path)])
        args = parser.parse_args()
        with pytest.raises(SystemExit):
            args.func(args)
        capture = capsys.readouterr()
        assert "does not exist" in capture.err
        assert str(path.parent) in capture.err

    def test_create_fails_when_markers_in_snakebids_version(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        capsys: pytest.CaptureFixture[str],
    ):
        mocker.patch.object(
            sys,
            "argv",
            ["snakebids", "create", "name", "--snakebids-version", "... ; markers"],
        )
        args = parser.parse_args()
        with pytest.raises(SystemExit):
            args.func(args)
        capture = capsys.readouterr()
        assert "may not specify markers" in capture.err

    def test_create_fails_when_get_spec_has_at_sign(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        capsys: pytest.CaptureFixture[str],
    ):
        mocker.patch.object(
            sys,
            "argv",
            ["snakebids", "create", "name", "--snakebids-version", "@ git+...@...@..."],
        )
        args = parser.parse_args()
        with pytest.raises(SystemExit):
            args.func(args)
        capture = capsys.readouterr()
        assert "in git requirement" in capture.err

    def test_create_fails_when_snakebids_version_specifies_extras(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        capsys: pytest.CaptureFixture[str],
    ):
        mocker.patch.object(
            sys,
            "argv",
            ["snakebids", "create", "name", "--snakebids-version", "[...] ..."],
        )
        args = parser.parse_args()
        with pytest.raises(SystemExit):
            args.func(args)
        capture = capsys.readouterr()
        assert "may not specify extras" in capture.err

    @given(
        name=st.from_regex(r"^[a-zA-Z_][a-zA-Z_0-9]*$"),
        version=st.text(st.characters(blacklist_characters=["@", ";", "["])).filter(
            lambda s: not s.startswith("-")
        )
        | st.none(),
    )
    @allow_function_scoped
    def test_create_calls_copier_correctly(
        self,
        parser: ArgumentParser,
        mocker: MockerFixture,
        name: str,
        version: str | None,
        tmp_path: Path,
        capsys: pytest.CaptureFixture[str],
    ):
        run_copy = mocker.patch.object(copier, "run_copy")
        path = tmp_path / name
        args = ["snakebids", "create", str(path)]
        if version is not None:
            args.extend(["--snakebids-version", version])
        mocker.patch.object(sys, "argv", args)
        args = parser.parse_args()
        args.func(args)
        capture = capsys.readouterr()
        assert "Creating Snakebids app at" in capture.err
        run_copy.assert_called_once_with(
            str(Path(itx.first(snakebids.__path__), "project_template")),
            path,
            data={
                "app_full_name": name,
                **({"snakebids_version": version} if version is not None else {}),
            },
            unsafe=True,
        )

    def test_create_succeeds(self, parser: ArgumentParser, mocker: MockerFixture):
        mocker.patch.object(sys, "argv", ["snakebids", "create"])
        assert isinstance(parser.parse_args(), Namespace)

    def test_boutiques_succeeds(self, parser: ArgumentParser, mocker: MockerFixture):
        mocker.patch.object(sys, "argv", ["snakebids", "boutiques", "test.json"])
        assert isinstance(parser.parse_args(), Namespace)


def test_boutiques_descriptor():
    with pytest.raises(SystemExit):
        create_descriptor(argparse.Namespace())



"""Tools to generate a Snakemake-based BIDS app.

This legacy module once had the core snakebids bidsapp implementation, but now is a
simple wrapper around :mod:`snakebids.bidsapp` with the
:class:`~snakebids.plugins.SnakemakeBidsApp` plugin. For new apps, this functionality
can be more flexibly implemented with:

.. code-block:: python

    from snakebids import bidsapp, plugins

    bidsapp.app([plugins.SnakemakeBidsApp(...)])

"""

from __future__ import annotations

import logging
import warnings
from collections.abc import Callable
from os import PathLike
from pathlib import Path
from typing import Any

import attrs
import boutiques.creator as bc  # type: ignore

from snakebids import bidsapp
from snakebids import plugins as sb_plugins
from snakebids.bidsapp.run import _Runner

logger = logging.Logger(__name__)

PluginList = list[Callable[["SnakeBidsApp"], "None | SnakeBidsApp"]]


@attrs.define
class SnakeBidsApp:
    """Snakebids app with config and arguments.

    Parameters
    ----------
    snakemake_dir : str | Path
        Root directory of the snakebids app, containing the config file and workflow
        files.
    plugins
        List of plugins to be registered.

        See :ref:`using-plugins` for more info.
    skip_parse_args
        DEPRECATED: no-op.
    parser
        DEPRECATED: no-op. (Historic: Parser including only the arguments
        specific to this Snakebids app, as specified in the config file. By
        default, it will use `create_parser()` from `cli.py`)
    configfile_path
        Relative path to config file (relative to snakemake_dir). By default,
        autocalculates based on snamake_dir
    snakefile_path
        Absolute path to the input Snakefile. By default, autocalculates based on
        snakemake_dir::

            join(snakemake_dir, snakefile_path)
    config
        DEPRECATED: no-op. (Historic: Contains all the configuration variables
        parsed from the config file and generated during the initialization of
        the SnakeBidsApp.)
    args
        DEPRECATED: no-op. (Historic: Arguments to use when running the app. By
        default, generated using the parser attribute, autopopulated with args
        from `config.py`)
    version
        DEPRECATED: no-op, use version plugin instead
    """

    snakemake_dir: Path = attrs.field(converter=Path)
    plugins: PluginList = attrs.Factory(PluginList)
    skip_parse_args: bool = False
    _parser: Any = attrs.field(default=None, alias="parser")
    configfile_path: Path | None = attrs.field(
        default=None, converter=attrs.converters.optional(Path)
    )
    snakefile_path: Path | None = attrs.field(
        default=None, converter=attrs.converters.optional(Path)
    )
    _config: Any = attrs.field(default=None, alias="config")
    version: str | None = None
    args: Any = None

    _app_holder: _Runner | None = attrs.field(default=None, init=False)

    @property
    def _app(self):
        if self._app_holder is None:
            self._check_deprecations()

            self._app_holder = bidsapp.app(
                [sb_plugins.SnakemakeBidsApp(**self._get_args()), *self.plugins],
                description="Snakebids helps build BIDS Apps with Snakemake",
            )
        return self._app_holder

    def _check_deprecations(self):
        if self._parser is not None:
            msg = (
                "`SnakeBidsApp.parser` is deprecated and no longer has any effect. To "
                "modify the parser, use the new `bidsapp` module."
            )
            warnings.warn(msg, stacklevel=3)
        if self._config is not None:
            msg = (
                "`SnakeBidsApp.config` is deprecated and no longer has any effect. To "
                "modify the config, use the new `bidsapp` module."
            )
            warnings.warn(msg, stacklevel=3)
        if self.args is not None:
            msg = (
                "`SnakeBidsApp.args` is deprecated and no longer has any effect. To "
                "modify CLI arguments, use the new `bidsapp` module."
            )
            warnings.warn(msg, stacklevel=3)
        if self.version is not None:
            msg = (
                "`SnakeBidsApp.version` is deprecated and no longer has any effect. To "
                "explicitly set the app version, use the `snakebids.plugins.Version` "
                "plugin"
            )
            warnings.warn(msg, stacklevel=3)

    @property
    def config(self):
        """Get config dict (before arguments are parsed)."""
        return self._app.build_parser().config

    @property
    def parser(self):
        """Get parser."""
        return self._app.build_parser().parser

    def _get_args(self):
        args: dict[str, Any] = {}
        args["snakemake_dir"] = self.snakemake_dir
        if self.configfile_path is not None:
            args["configfile_path"] = self.configfile_path
        if self.snakefile_path is not None:
            args["snakefile_path"] = self.snakefile_path
        return args

    def run_snakemake(self) -> None:
        """Run snakemake with the given config, after applying plugins."""
        self._app.run()

    def create_descriptor(self, out_file: PathLike[str] | str) -> None:
        """Generate a boutiques descriptor for this Snakebids app."""
        new_descriptor = bc.CreateDescriptor(  # type: ignore
            self.parser, execname="run.py"
        )
        new_descriptor.save(out_file)  # type: ignore



"""
This type stub file was generated by pyright.
"""

if False: ...
anchor_attrib = ...

class Anchor:
    __slots__ = ...
    attrib = ...
    def __init__(self) -> None: ...
    def __repr__(self) -> Any: ...



from .datasets import (
    BidsComponent,
    BidsComponentRow,
    BidsDataset,
    BidsDatasetDict,
    BidsPartialComponent,
)
from .filtering import (
    filter_list,
    get_filtered_ziplist_index,
)
from .input_generation import (
    generate_inputs,
    get_wildcard_constraints,
    write_derivative_json,
)

__all__ = [
    "BidsComponent",
    "BidsComponentRow",
    "BidsDataset",
    "BidsDatasetDict",
    "BidsPartialComponent",
    "filter_list",
    "generate_inputs",
    "get_filtered_ziplist_index",
    "get_wildcard_constraints",
    "write_derivative_json",
]



from collections.abc import Callable
from typing import Any

class UserProperty[T]:
    def __init__(self, method: Callable[[Any], T], /) -> None: ...
    def __get__(self, obj: Any, objtype: type[Any] = ...) -> T: ...
